r"""
Super Hopf algebras with basis
"""
# ****************************************************************************
#  Copyright (C) 2015 Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************
from sage.misc.lazy_attribute import lazy_attribute
from sage.categories.super_modules import SuperModulesCategory


class SuperHopfAlgebrasWithBasis(SuperModulesCategory):
    """
    The category of super Hopf algebras with a distinguished basis.

    EXAMPLES::

        sage: C = HopfAlgebras(ZZ).WithBasis().Super(); C
        Category of super hopf algebras with basis over Integer Ring
        sage: sorted(C.super_categories(), key=str)
        [Category of super algebras with basis over Integer Ring,
         Category of super coalgebras with basis over Integer Ring,
         Category of super hopf algebras over Integer Ring]

    TESTS::

        sage: C = HopfAlgebras(ZZ).WithBasis().Super()
        sage: TestSuite(C).run()
    """
    class ParentMethods:
        @lazy_attribute
        def antipode(self):
            """
            The antipode of this Hopf algebra.

            If :meth:`.antipode_basis` is available, this constructs the
            antipode morphism from ``self`` to ``self`` by extending it by
            linearity. Otherwise, :meth:`self.antipode_by_coercion` is used,
            if available.

            EXAMPLES::

                sage: A = SteenrodAlgebra(7)
                sage: a = A.an_element()
                sage: a, A.antipode(a)
                (6 Q_1 Q_3 P(2,1), Q_1 Q_3 P(2,1))

            TESTS::

                sage: E.<x,y> = ExteriorAlgebra(QQ)
                sage: [b.antipode() for b in E.basis()]
                [1, -x, -y, x*y]
            """
            if self.antipode_on_basis is not NotImplemented:
                # Should give the information that this is an anti-morphism of algebra
                return self._module_morphism(self.antipode_on_basis, codomain = self)
            elif hasattr(self, "antipode_by_coercion"):
                return self.antipode_by_coercion

        def _test_antipode(self, **options):
            r"""
            Test the antipode.

            An *antipode* `S` of a (super) Hopf algebra is a linear
            endomorphism of the Hopf algebra that satisfies the
            following conditions (see :wikipedia:`HopfAlgebra`).

            - If `\mu` and `\Delta` denote the product and coproduct of the
              Hopf algebra, respectively, then `S` satisfies

              .. MATH::

                  \mu \circ (S \tensor 1) \circ \Delta = unit \circ counit
                  \mu \circ (1 \tensor S) \circ \Delta = unit \circ counit

            - `S` is an *anti*-homomorphism:

               .. MATH::

                   S(ab) = (-1)^{\deg a \deg b} S(b) S(a)

               for homogeneous `a` and `b`.

            These properties are tested on :meth:`some_elements`.

            TESTS::

                sage: A = SteenrodAlgebra(7)
                sage: A._test_antipode()  # long time
            """
            tester = self._tester(**options)

            S = self.antipode

            IS = lambda x: self.sum(c * self.monomial(t1) * S(self.monomial(t2))
                                for ((t1, t2), c) in x.coproduct())

            SI = lambda x: self.sum(c * S(self.monomial(t1)) * self.monomial(t2)
                                for ((t1, t2), c) in x.coproduct())

            for x in tester.some_elements():
                x_even = x.even_component()
                x_odd = x.odd_component()
                for y in tester.some_elements():
                    y_even = y.even_component()
                    y_odd = y.odd_component()

                    # The antipode is a graded anti-homomorphism.
                    tester.assertEqual(S(x_even) * S(y_even),
                                       S(y_even * x_even))
                    tester.assertEqual(S(x_even) * S(y_odd),
                                       S(y_odd * x_even))
                    tester.assertEqual(S(x_odd) * S(y_even),
                                       S(y_even * x_odd))
                    tester.assertEqual(S(x_odd) * S(y_odd),
                                       -S(y_odd * x_odd))

                # mu * (S # I) * delta == counit * unit
                tester.assertEqual(SI(x), self.counit(x) * self.one())

                # mu * (I # S) * delta == counit * unit
                tester.assertEqual(IS(x), self.counit(x) * self.one())




r"""
Group, ring, etc. actions on objects

The terminology and notation used is suggestive of groups acting on sets,
but this framework can be used for modules, algebras, etc.

A group action $G \times S \rightarrow S$ is a functor from $G$ to Sets.

.. WARNING::

    An :class:`Action` object only keeps a weak reference to the underlying set
    which is acted upon. This decision was made in :trac:`715` in order to
    allow garbage collection within the coercion framework (this is where
    actions are mainly used) and avoid memory leaks.

    ::

        sage: from sage.categories.action import Action
        sage: class P: pass
        sage: A = Action(P(),P())
        sage: import gc
        sage: _ = gc.collect()
        sage: A
        <repr(<sage.categories.action.Action at 0x...>) failed: RuntimeError: This action acted on a set that became garbage collected>

    To avoid garbage collection of the underlying set, it is sufficient to
    create a strong reference to it before the action is created.

    ::

        sage: _ = gc.collect()
        sage: from sage.categories.action import Action
        sage: class P: pass
        sage: q = P()
        sage: A = Action(P(),q)
        sage: gc.collect()
        0
        sage: A
        Left action by <__main__.P ... at ...> on <__main__.P ... at ...>

AUTHOR:

- Robert Bradshaw: initial version
"""

#*****************************************************************************
#       Copyright (C) 2007 Robert Bradshaw <robertwb@math.washington.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from cpython.tuple cimport PyTuple_GET_ITEM

from .functor cimport Functor
from .morphism cimport Morphism
from .map cimport Map
from sage.structure.element cimport parent
from sage.structure.parent cimport Parent

from . import homset
import sage.structure.element
from weakref import ref
from sage.misc.constant_function import ConstantFunction


cdef inline category(x):
    try:
        return x.category()
    except AttributeError:
        from sage.categories.objects import Objects
        return Objects()


cdef class Action(Functor):
    """
    The action of ``G`` on ``S``.

    INPUT:

    - ``G`` -- a parent or Python type

    - ``S`` -- a parent or Python type

    - ``is_left`` -- (boolean, default: ``True``) whether elements of
      ``G`` are on the left

    - ``op`` -- (default: ``None``) operation. This is not used by
      :class:`Action` itself, but other classes may use it
    """
    def __init__(self, G, S, is_left=True, op=None):
        from .groupoid import Groupoid
        Functor.__init__(self, Groupoid(G), category(S))
        self.G = G
        self.US = ref(S)
        self._is_left = is_left
        self.op = op

    def _apply_functor(self, x):
        return self(x)

    def __reduce__(self):
        """
        Used in pickling.

        .. WARNING::

            If you change the signature of the ``__init__`` for a subclass,
            you must override this method as well.

        TESTS:

        Check that this action can be pickled (:trac:`29031`)::

            sage: P = QQ['x']
            sage: R = (ZZ['x'])['y']
            sage: A = R.get_action(P, operator.mul, True)
            sage: loads(dumps(A)) is not None
            True
        """
        return (type(self), (self.G, self.underlying_set(), self._is_left, self.op))

    def __call__(self, *args):
        """
        Let this action act.

        For a left action, ``action(a, b)`` lets ``a`` act on ``b``.
        For a right action, ``action(a, b)`` lets ``b`` act on ``a``.

        If needed, ``a`` and ``b`` are converted to the correct parent.

        .. SEEALSO::

            :meth:`act` which lets you pass the acting and acted-on
            elements directly.

        EXAMPLES::

            sage: R.<x> = ZZ []
            sage: from sage.structure.coerce_actions import IntegerMulAction
            sage: A = IntegerMulAction(ZZ, R, True)   # Left action
            sage: A(5, x)
            5*x
            sage: A(int(5), x)
            5*x
            sage: A(x, 5)
            Traceback (most recent call last):
            ...
            TypeError: not a constant polynomial
            sage: A = IntegerMulAction(ZZ, R, False)  # Right action
            sage: A(x, 5)
            5*x
            sage: A(x, int(5))
            5*x
            sage: A(5, x)
            Traceback (most recent call last):
            ...
            TypeError: not a constant polynomial
        """
        if len(args) == 2:
            # Normal case, called with (g, x) or (x, g) as arguments
            #
            # For a left action:  g = args[0] and x = args[1]
            # For a right action: g = args[1] and x = args[0]
            g = <object>PyTuple_GET_ITEM(args, 1 - self._is_left)
            x = <object>PyTuple_GET_ITEM(args, self._is_left)
            return self._act_convert(g, x)
        elif len(args) == 1:
            g = <object>PyTuple_GET_ITEM(args, 0)
            if g in self.G:
                return ActionEndomorphism(self, self.G(g))
            elif g == self.G:
                return self.underlying_set()
            else:
                raise TypeError("%s not an element of %s" % (g, self.G))
        else:
            raise TypeError("actions should be called with 1 or 2 arguments")

    cdef _act_convert(self, g, x):
        """
        Let ``g`` act on ``x`` under this action, converting ``g``
        and ``x`` to the correct parents first.
        """
        U = self.underlying_set()
        if parent(g) is not self.G:
            g = self.G(g)
        if parent(x) is not U:
            x = U(x)
        return self._act_(g, x)

    cpdef _act_(self, g, x):
        """
        Let ``g`` act on ``x`` under this action.

        Regardless of whether this is a left or right action, the acting
        element comes first.

        INPUT:

        - ``g`` -- an object with parent ``self.G``.

        - ``x`` -- an object with parent ``self.US()``.

        .. WARNING::

            This is meant to be a fast internal function, so the
            conditions on the input are not checked!
        """
        raise NotImplementedError(f"action for {type(self)} not implemented")

    def act(self, g, x):
        """
        This is a consistent interface for acting on ``x`` by ``g``,
        regardless of whether it's a left or right action.

        If needed, ``g`` and ``x`` are converted to the correct parent.

        EXAMPLES::

            sage: R.<x> = ZZ []
            sage: from sage.structure.coerce_actions import IntegerMulAction
            sage: A = IntegerMulAction(ZZ, R, True)   # Left action
            sage: A.act(5, x)
            5*x
            sage: A.act(int(5), x)
            5*x
            sage: A = IntegerMulAction(ZZ, R, False)  # Right action
            sage: A.act(5, x)
            5*x
            sage: A.act(int(5), x)
            5*x
        """
        return self._act_convert(g, x)

    def __invert__(self):
        return InverseAction(self)

    def is_left(self):
        return self._is_left

    def _repr_(self):
        side = "Left" if self._is_left else "Right"
        return "%s %s by %r on %r"%(side, self._repr_name_(), self.G,
                                    self.underlying_set())

    def _repr_name_(self):
        return "action"

    def actor(self):
        return self.G

    cdef underlying_set(self):
        """
        The set on which the actor acts (it is not necessarily the codomain of
        the action).

        .. NOTE::

            Since this is a cdef'ed method, we can only provide an indirect doctest.

        EXAMPLES::

            sage: P = QQ['x']
            sage: R = (ZZ['x'])['y']
            sage: A = R.get_action(P,operator.mul,True)
            sage: A                 # indirect doctest
            Right scalar multiplication by Univariate Polynomial Ring in x over
            Rational Field on Univariate Polynomial Ring in y over Univariate
            Polynomial Ring in x over Integer Ring

        In this example, the underlying set is the ring ``R``. This is the same
        as the left domain, which is different from the codomain of the action::

            sage: A.codomain()
            Univariate Polynomial Ring in y over Univariate Polynomial Ring in x over Rational Field
            sage: A.codomain() == R
            False
            sage: A.left_domain() is R
            True

        By :trac:`715`, there is only a weak reference to the underlying set.
        Hence, the underlying set may be garbage collected, even when the
        action is still alive. This may result in a runtime error, as follows::

            sage: from sage.categories.action import Action
            sage: class P: pass
            sage: p = P()
            sage: q = P()
            sage: A = Action(p,q)
            sage: A
            Left action by <__main__.P ... at ...> on <__main__.P ... at ...>
            sage: del q
            sage: import gc
            sage: _ = gc.collect()
            sage: A
            <repr(<sage.categories.action.Action at 0x...>) failed: RuntimeError: This action acted on a set that became garbage collected>
        """
        S = self.US()
        if S is None:
            raise RuntimeError("This action acted on a set that became garbage collected")
        return S

    def codomain(self):
       return self.underlying_set()

    def domain(self):
        return self.underlying_set()

    def left_domain(self):
        if self._is_left:
            return self.G
        else:
            return self.domain()

    def right_domain(self):
        if self._is_left:
            return self.domain()
        else:
            return self.G

    def operation(self):
        return self.op


cdef class InverseAction(Action):
    """
    An action that acts as the inverse of the given action.

    EXAMPLES::

        sage: V = QQ^3
        sage: v = V((1, 2, 3))
        sage: cm = get_coercion_model()

        sage: a = cm.get_action(V, QQ, operator.mul)
        sage: a
        Right scalar multiplication by Rational Field on Vector space of dimension 3 over Rational Field
        sage: ~a
        Right inverse action by Rational Field on Vector space of dimension 3 over Rational Field
        sage: (~a)(v, 1/3)
        (3, 6, 9)

        sage: b = cm.get_action(QQ, V, operator.mul)
        sage: b
        Left scalar multiplication by Rational Field on Vector space of dimension 3 over Rational Field
        sage: ~b
        Left inverse action by Rational Field on Vector space of dimension 3 over Rational Field
        sage: (~b)(1/3, v)
        (3, 6, 9)

        sage: c = cm.get_action(ZZ, list, operator.mul)
        sage: c
        Left action by Integer Ring on <... 'list'>
        sage: ~c
        Traceback (most recent call last):
        ...
        TypeError: no inverse defined for Left action by Integer Ring on <... 'list'>

    TESTS:

        sage: x = polygen(QQ,'x')
        sage: a = 2*x^2+2; a
        2*x^2 + 2
        sage: a / 2
        x^2 + 1
        sage: a /= 2
        sage: a
        x^2 + 1
    """
    def __init__(self, Action action):
        G = action.G
        try:
            from sage.groups.group import is_Group
            # We must be in the case that parent(~a) == parent(a)
            # so we can invert in _call_ code below.
            if (is_Group(G) and G.is_multiplicative()) or G.is_field():
                Action.__init__(self, G, action.underlying_set(), action._is_left)
                self._action = action
                return
        except (AttributeError, NotImplementedError):
            pass
        raise TypeError(f"no inverse defined for {action!r}")

    def __reduce__(self):
        """
        Used in pickling.

        TESTS:

        Check that this action can be pickled (:trac:`29031`)::

            sage: V = QQ^3
            sage: v = V((1, 2, 3))
            sage: cm = get_coercion_model()
            sage: a = cm.get_action(V, QQ, operator.mul)
            sage: loads(dumps(~a)) is not None
            True
        """
        return (type(self), (self._action,))

    cpdef _act_(self, g, x):
        if self.S_precomposition is not None:
            x = self.S_precomposition(x)
        return self._action._act_(~g, x)

    def codomain(self):
        return self._action.codomain()

    def __invert__(self):
        return self._action

    def _repr_name_(self):
        return "inverse action"


cdef class PrecomposedAction(Action):
    """
    A precomposed action first applies given maps, and then applying an action
    to the return values of the maps.

    EXAMPLES:

    We demonstrate that an example discussed on :trac:`14711` did not become a
    problem::

        sage: E = ModularSymbols(11).2
        sage: s = E.modular_symbol_rep()
        sage: del E,s
        sage: import gc
        sage: _ = gc.collect()
        sage: E = ModularSymbols(11).2
        sage: v = E.manin_symbol_rep()
        sage: c,x = v[0]
        sage: y = x.modular_symbol_rep()
        sage: coercion_model.get_action(QQ, parent(y), op=operator.mul)
        Left scalar multiplication by Rational Field on Abelian Group of all Formal Finite Sums over Rational Field
        with precomposition on right by Coercion map:
          From: Abelian Group of all Formal Finite Sums over Integer Ring
          To:   Abelian Group of all Formal Finite Sums over Rational Field
    """
    def __init__(self, Action action, Map left_precomposition, Map right_precomposition):
        left = action.left_domain()
        right = action.right_domain()
        US = action.underlying_set()
        cdef Parent lco, rco
        if left_precomposition is not None:
            lco = left_precomposition._codomain
            if lco is not left:
                left_precomposition = homset.Hom(lco, left).natural_map() * left_precomposition
            left = left_precomposition.domain()
        if right_precomposition is not None:
            rco = right_precomposition._codomain
            if rco is not right:
              right_precomposition = homset.Hom(rco, right).natural_map() * right_precomposition
            right = right_precomposition.domain()
        if action._is_left:
            Action.__init__(self, left, US, 1)
        else:
            Action.__init__(self, right, US, 0)
        self._action = action
        if self._is_left:
            self.G_precomposition = left_precomposition
            self.S_precomposition = right_precomposition
        else:
            self.G_precomposition = right_precomposition
            self.S_precomposition = left_precomposition

    def __reduce__(self):
        """
        Used in pickling.

        TESTS:

        Check that this action can be pickled (:trac:`29031`)::

            sage: E = ModularSymbols(11).2
            sage: v = E.manin_symbol_rep()
            sage: c,x = v[0]
            sage: y = x.modular_symbol_rep()
            sage: act = coercion_model.get_action(QQ, parent(y), op=operator.mul)
            sage: loads(dumps(act)) is not None
            True
        """
        return (type(self), (self._action, self.G_precomposition, self.S_precomposition))

    cpdef _act_(self, g, x):
        if self.G_precomposition is not None:
            g = self.G_precomposition._call_(g)
        if self.S_precomposition is not None:
            x = self.S_precomposition._call_(x)
        return self._action._act_(g, x)

    def domain(self):
        if self.S_precomposition is not None:
            return self.S_precomposition.domain()
        else:
            return self._action.domain()

    def codomain(self):
        return self._action.codomain()

    @property
    def left_precomposition(self):
        """
        The left map to precompose with, or ``None`` if there is no
        left precomposition map.
        """
        if self._is_left:
            return self.G_precomposition
        else:
            return self.S_precomposition

    @property
    def right_precomposition(self):
        """
        The right map to precompose with, or ``None`` if there is no
        right precomposition map.
        """
        if self._is_left:
            return self.S_precomposition
        else:
            return self.G_precomposition

    def __invert__(self):
        return PrecomposedAction(~self._action, self.left_precomposition, self.right_precomposition)

    def _repr_(self):
        s = repr(self._action)
        if self.left_precomposition is not None:
            s += "\nwith precomposition on left by %s" % self.left_precomposition._default_repr_()
        if self.right_precomposition is not None:
            s += "\nwith precomposition on right by %s" % self.right_precomposition._default_repr_()
        return s


cdef class ActionEndomorphism(Morphism):
    """
    The endomorphism defined by the action of one element.

    EXAMPLES::

        sage: A = ZZ['x'].get_action(QQ, self_on_left=False, op=operator.mul)
        sage: A
        Left scalar multiplication by Rational Field on Univariate Polynomial
        Ring in x over Integer Ring
        sage: A(1/2)
        Action of 1/2 on Univariate Polynomial Ring in x over Integer Ring
        under Left scalar multiplication by Rational Field on Univariate
        Polynomial Ring in x over Integer Ring.
    """
    def __init__(self, Action action, g):
        Morphism.__init__(self, homset.Hom(action.underlying_set(),
                                           action.underlying_set()))
        self._action = action
        self._g = g

    cdef dict _extra_slots(self):
        """
        Helper for pickling and copying.

        TESTS::

            sage: P.<x> = ZZ[]
            sage: A = P.get_action(QQ, self_on_left=False, op=operator.mul)
            sage: phi = A(1/2)
            sage: psi = copy(phi)  # indirect doctest
            sage: psi
            Action of 1/2 on Univariate Polynomial Ring in x over
            Integer Ring under Left scalar multiplication by Rational
            Field on Univariate Polynomial Ring in x over Integer Ring.
            sage: psi(x) == phi(x)
            True
        """
        slots = Morphism._extra_slots(self)
        slots['_action'] = self._action
        slots['_g'] = self._g
        return slots

    cdef _update_slots(self, dict _slots):
        """
        Helper for pickling and copying.

        TESTS::

            sage: P.<x> = ZZ[]
            sage: A = P.get_action(QQ, self_on_left=False, op=operator.mul)
            sage: phi = A(1/2)
            sage: psi = copy(phi)  # indirect doctest
            sage: psi
            Action of 1/2 on Univariate Polynomial Ring in x over
            Integer Ring under Left scalar multiplication by Rational
            Field on Univariate Polynomial Ring in x over Integer Ring.
            sage: psi(x) == phi(x)
            True
        """
        self._action = _slots['_action']
        self._g = _slots['_g']
        Morphism._update_slots(self, _slots)

    cpdef Element _call_(self, x):
        return self._action._act_(self._g, x)

    def _repr_(self):
        return "Action of %s on %s under %s."%(self._g,
                                               self._action.underlying_set(), self._action)

    def __mul__(left, right):
        cdef ActionEndomorphism left_c, right_c
        if isinstance(left, ActionEndomorphism) and isinstance(right, ActionEndomorphism):
            left_c = left
            right_c = right
            if left_c._action is right_c._action:
                if left_c._action._is_left:
                    return ActionEndomorphism(left_c._action, left_c._g * right_c._g)
                else:
                    return ActionEndomorphism(left_c._action, right_c._g * left_c._g)
        return Morphism.__mul__(left, right)

    def __invert__(self):
        inv_g = ~self._g
        if parent(inv_g) is parent(self._g):
            return ActionEndomorphism(self._action, inv_g)
        else:
            return (~self._action)(self._g)



"""
Cython Functions for Timing Clone Protocol
"""
#*****************************************************************************
#  Copyright (C) 2009-2010 Florent Hivert <Florent.Hivert@univ-rouen.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.structure.list_clone cimport ClonableArray


#####################################################################
######                    Timings functions                    ######
#####################################################################
cpdef ClonableArray cy_add1_internal(ClonableArray bla):
    """
    TESTS::

        sage: from sage.structure.list_clone_timings_cy import *
        sage: from sage.structure.list_clone_timings import *
        sage: cy_add1_internal(IncreasingArrays()([1,4,5]))
        [2, 5, 6]
    """
    cdef int i
    cdef list lst
    cdef ClonableArray blo
    blo = bla.__copy__()
    lst = blo._get_list()
    for i in range(len(lst)): lst[i] += 1
    blo.set_immutable()
    blo.check()
    return blo


cpdef ClonableArray cy_add1_immutable(ClonableArray bla):
    """
    TESTS::

        sage: from sage.structure.list_clone_timings import *
        sage: from sage.structure.list_clone_timings_cy import *
        sage: cy_add1_immutable(IncreasingArrays()([1,4,5]))
        [2, 5, 6]
    """
    cdef int i
    cdef list lbla
    lbla = bla[:]
    for i in range(len(lbla)): lbla[i] += 1
    return bla.__class__(bla._parent, lbla)

cpdef ClonableArray cy_add1_mutable(ClonableArray bla):
    """
    TESTS::

        sage: from sage.structure.list_clone_timings import *
        sage: from sage.structure.list_clone_timings_cy import *
        sage: cy_add1_mutable(IncreasingArrays()([1,4,5]))
        [2, 5, 6]
    """
    cdef int i
    cdef ClonableArray blo
    blo = bla.__copy__()
    for i in range(len(blo)):
        blo._setitem(i, blo._getitem(i)+1)
    blo.set_immutable()
    blo.check()
    return blo


cpdef ClonableArray cy_add1_with(ClonableArray bla):
    """
    TESTS::

        sage: from sage.structure.list_clone_timings import *
        sage: from sage.structure.list_clone_timings_cy import *
        sage: cy_add1_with(IncreasingArrays()([1,4,5]))
        [2, 5, 6]
    """
    cdef int i
    cdef ClonableArray blo
    with bla.__copy__() as blo:
        for i in range(len(blo)):
            blo._setitem(i, blo._getitem(i)+1)
    return blo



from sage.categories.map cimport Map


cdef class DefaultConvertMap(Map):
    pass


cdef class DefaultConvertMap_unique(DefaultConvertMap):
    pass


cdef class NamedConvertMap(Map):
    cdef readonly method_name


cdef class TryMap(Map):
    cdef Map _map_p
    cdef Map _map_b
    cdef _error_types


cdef class CallableConvertMap(Map):
    cdef bint _parent_as_first_arg
    cdef _func


cdef Map CCallableConvertMap(domain, codomain, void* func, name)



from .function_mangling cimport ArgumentFixer

cpdef dict_key(o)
cpdef cache_key(o)

cdef class CachedFunction(object):
    cdef public str __name__
    cdef public str __module__
    cdef ArgumentFixer _argument_fixer
    cdef public f
    cdef public cache  # not always of type <dict>
    cdef bint is_classmethod
    cdef int argfix_init(self) except -1
    cdef get_key_args_kwds(self, tuple args, dict kwds)
    cdef fix_args_kwds(self, tuple args, dict kwds)
    cdef empty_key
    cdef key
    cdef bint do_pickle

cdef class CachedMethod(object):
    cdef str _cache_name
    cdef public str __name__
    cdef public str __module__
    cdef CachedFunction _cachedfunc
    cdef Py_ssize_t nargs
    cpdef _get_instance_cache(self, inst)

cdef class CacheDict(dict):
    pass

cdef class CachedInParentMethod(CachedMethod):
    pass

cdef class CachedMethodCaller(CachedFunction):
    cdef public _instance
    cdef public CachedMethod _cachedmethod

cdef class CachedMethodCallerNoArgs(CachedFunction):
    cdef public _instance

cdef class GloballyCachedMethodCaller(CachedMethodCaller):
    pass



# sage.cpython is an ordinary package, not a namespace package.

# This package is imported very early, which is why workarounds/monkey-patching
# are done in this file.

# Make sure that the correct zlib library is loaded. This is needed
# to prevent the system zlib to be loaded instead of the Sage one.
# See https://trac.sagemath.org/ticket/23122
import zlib as _zlib
del _zlib

# Monkey-patch ExtensionFileLoader to allow IPython to find the sources
# of Cython files. See https://trac.sagemath.org/ticket/24681
from importlib.machinery import ExtensionFileLoader as _ExtensionFileLoader
del _ExtensionFileLoader.get_source
del _ExtensionFileLoader

# Work around a Cygwin-specific bug caused by sqlite3; see
# https://trac.sagemath.org/ticket/30157 and the docstring for
# fix_for_ticket_30157
# Here we monkey-patch the sqlite3 module to ensure the fix is
# applied the very first time a connection is made to a sqlite3
# database
import sys as _sys
if _sys.platform == 'cygwin':
    def _patch_sqlite3():
        try:
            from sage.misc.sage_ostools import fix_for_ticket_30157
        except ImportError:
            # The module might not have been re-built yet; don't worry about it
            # then
            return

        import sqlite3
        import functools
        orig_sqlite3_connect = sqlite3.connect

        @functools.wraps(orig_sqlite3_connect)
        def connect(*args, **kwargs):
            if fix_for_ticket_30157():
                raise RuntimeError(
                    'patch for Trac ticket #30157 failed; please report this '
                    'bug to https://trac.sagemath.org')

            # Undo the monkey-patch
            try:
                return orig_sqlite3_connect(*args, **kwargs)
            finally:
                sqlite3.connect = orig_sqlite3_connect

        sqlite3.connect = connect

    _patch_sqlite3()
    del _patch_sqlite3
    del _sys



"""
Metaclasses for Cython extension types

Cython does not support metaclasses, but this module can be used to
implement metaclasses for extension types.

.. WARNING::

    This module has many caveats and you can easily get segfaults if you
    make a mistake. It relies on undocumented Python and Cython
    behaviour, so things might break in future versions.

How to use
==========

To enable this metaclass mechanism, you need to put
``cimport sage.cpython.cython_metaclass`` in your module (in the ``.pxd``
file if you are using one).

In the extension type (a.k.a. ``cdef class``) for which you want to
define a metaclass, define a method ``__getmetaclass__`` with a single
unused argument. This method should return a type to be used as
metaclass:

.. code-block:: cython

    cimport sage.cpython.cython_metaclass
    cdef class MyCustomType(object):
        def __getmetaclass__(_):
            from foo import MyMetaclass
            return MyMetaclass

.. WARNING::

    ``__getmetaclass__`` must be defined as an ordinary method taking a
    single argument, but this argument should not be used in the
    method (it will be ``None``).

When a type ``cls`` is being constructed with metaclass ``meta``,
then ``meta.__init__(cls, None, None, None)`` is called from Cython.
In Python, this would be ``meta.__init__(cls, name, bases, dict)``.

.. WARNING::

    The ``__getmetaclass__`` method is called while the type is being
    created during the import of the module. Therefore,
    ``__getmetaclass__`` should not refer to any global objects,
    including the type being created or other types defined or imported
    in the module (unless you are very careful). Note that non-imported
    ``cdef`` functions are not Python objects, so those are safe to call.

    The same warning applies to the ``__init__`` method of the
    metaclass.

.. WARNING::

    The ``__new__`` method of the metaclass (including the ``__cinit__``
    method for Cython extension types) is never called if you're using
    this from Cython. In particular, the metaclass cannot have any
    attributes or virtual methods.

EXAMPLES::

    sage: cython('''
    ....: cimport sage.cpython.cython_metaclass
    ....: cdef class MyCustomType(object):
    ....:     def __getmetaclass__(_):
    ....:         class MyMetaclass(type):
    ....:             def __init__(*args):
    ....:                 print("Calling MyMetaclass.__init__{}".format(args))
    ....:         return MyMetaclass
    ....:
    ....: cdef class MyDerivedType(MyCustomType):
    ....:     pass
    ....: ''')
    Calling MyMetaclass.__init__(<class '...MyCustomType'>, None, None, None)
    Calling MyMetaclass.__init__(<class '...MyDerivedType'>, None, None, None)
    sage: MyCustomType.__class__
    <class '...MyMetaclass'>
    sage: class MyPythonType(MyDerivedType):
    ....:     pass
    Calling MyMetaclass.__init__(<class '...MyPythonType'>, 'MyPythonType', (<class '...MyDerivedType'>,), {...})

Implementation
==============

All this is implemented by defining

.. code-block:: c

    #define PyTypeReady(t)  Sage_PyType_Ready(t)

and then implementing the function ``Sage_PyType_Ready(t)`` which first
calls ``PyType_Ready(t)`` and then handles the metaclass stuff.

TESTS:

Check that a proper exception is raised if ``__getmetaclass__``
returns a non-type::

    sage: cython('''
    ....: cimport sage.cpython.cython_metaclass
    ....: cdef class MyCustomType(object):
    ....:     def __getmetaclass__(_):
    ....:         return 2
    ....: ''')
    Traceback (most recent call last):
    ...
    TypeError: __getmetaclass__ did not return a type
"""

#*****************************************************************************
#       Copyright (C) 2015 Jeroen Demeyer <jdemeyer@cage.ugent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************



r"""
Generalized Coxeter Groups
"""
#*****************************************************************************
#  Copyright (C) 2016 Travis Scrimshaw <tscrim at ucdavis.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.categories.category_singleton import Category_singleton
from sage.categories.category_with_axiom import CategoryWithAxiom
from sage.categories.complex_reflection_or_generalized_coxeter_groups import ComplexReflectionOrGeneralizedCoxeterGroups

class GeneralizedCoxeterGroups(Category_singleton):
    r"""
    The category of generalized Coxeter groups.

    A generalized Coxeter group is a group with a presentation of
    the following form:

    .. MATH::

        \langle s_i \mid s_i^{p_i}, s_i s_j \cdots = s_j s_i \cdots \rangle,

    where `p_i > 1`, `i \in I`, and the factors in the braid relation
    occur `m_{ij} = m_{ji}` times for all `i \neq j \in I`.

    EXAMPLES::

        sage: from sage.categories.generalized_coxeter_groups import GeneralizedCoxeterGroups
        sage: C = GeneralizedCoxeterGroups(); C
        Category of generalized coxeter groups

    TESTS::

        sage: TestSuite(C).run()
    """
    @cached_method
    def super_categories(self):
        """
        EXAMPLES::

            sage: from sage.categories.generalized_coxeter_groups import GeneralizedCoxeterGroups
            sage: GeneralizedCoxeterGroups().super_categories()
            [Category of complex reflection or generalized coxeter groups]
        """
        return [ComplexReflectionOrGeneralizedCoxeterGroups()]

    def additional_structure(self):
        r"""
        Return ``None``.

        Indeed, all the structure generalized Coxeter groups have in
        addition to groups (simple reflections, ...) is already
        defined in the super category.

        .. SEEALSO:: :meth:`Category.additional_structure`

        EXAMPLES::

            sage: from sage.categories.generalized_coxeter_groups import GeneralizedCoxeterGroups
            sage: GeneralizedCoxeterGroups().additional_structure()
        """
        return None


    class Finite(CategoryWithAxiom):
        """
        The category of finite generalized Coxeter groups.
        """
        def extra_super_categories(self):
            """
            Implement that a finite generalized Coxeter group is a
            well-generated complex reflection group.

            EXAMPLES::

                sage: from sage.categories.generalized_coxeter_groups import GeneralizedCoxeterGroups
                sage: from sage.categories.complex_reflection_groups import ComplexReflectionGroups

                sage: Cat = GeneralizedCoxeterGroups().Finite()
                sage: Cat.extra_super_categories()
                [Category of well generated finite complex reflection groups]
                sage: Cat.is_subcategory(ComplexReflectionGroups().Finite().WellGenerated())
                True
            """
            from sage.categories.complex_reflection_groups import ComplexReflectionGroups
            return [ComplexReflectionGroups().Finite().WellGenerated()]




# A class that allows for a more efficient creation
# of attribute errors, so that raising them requires
# less time.
cdef class AttributeErrorMessage:
    cdef public cls
    cdef public name

cpdef getattr_from_other_class(self, cls, name)



"""
All imports for Jupyter
"""

from sage.all_cmdline import *

from .widgets_sagenb import (input_box, text_control, slider,
        range_slider, checkbox, selector, input_grid, color_selector)
from .interact import interact



r"""
Finite Dimensional Graded Lie Algebras With Basis

AUTHORS:

- Eero Hakavuori (2018-08-16): initial version
"""

# ****************************************************************************
#       Copyright (C) 2018 Eero Hakavuori <eero.hakavuori@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.categories.category_with_axiom import CategoryWithAxiom_over_base_ring

class FiniteDimensionalGradedLieAlgebrasWithBasis(CategoryWithAxiom_over_base_ring):
    r"""
    Category of finite dimensional graded Lie algebras with a basis.

    A grading of a Lie algebra `\mathfrak{g}` is a direct sum decomposition
    `\mathfrak{g} = \bigoplus_{i} V_i` such that `[V_i,V_j] \subset V_{i+j}`.

    EXAMPLES::

        sage: C = LieAlgebras(ZZ).WithBasis().FiniteDimensional().Graded(); C
        Category of finite dimensional graded lie algebras with basis over Integer Ring
        sage: C.super_categories()
        [Category of graded lie algebras with basis over Integer Ring,
         Category of finite dimensional lie algebras with basis over Integer Ring]

        sage: C is LieAlgebras(ZZ).WithBasis().FiniteDimensional().Graded()
        True

    TESTS::

        sage: C = LieAlgebras(QQ).FiniteDimensional().WithBasis().Graded()
        sage: TestSuite(C).run()
    """
    class ParentMethods:
        def _test_grading(self, **options):
            r"""
            Tests that the Lie bracket respects the grading.

            INPUT:

            - ``options`` -- any keyword arguments accepted by :meth:`_tester`.

            EXAMPLES::

                sage: C = LieAlgebras(QQ).WithBasis().Graded()
                sage: C = C.FiniteDimensional().Stratified().Nilpotent()
                sage: L = LieAlgebra(QQ, {('x','y'): {'z': 1}},
                ....:                nilpotent=True, category=C)
                sage: L._test_grading()
                sage: L = LieAlgebra(QQ, {('x','y'): {'x': 1}},
                ....:                nilpotent=True, category=C)
                sage: L._test_grading()
                Traceback (most recent call last):
                ...
                AssertionError: Lie bracket [x, y] has degree 1, not degree 2

            See the documentation for :class:`TestSuite` for more information.
            """
            tester = self._tester(**options)

            from sage.misc.misc import some_tuples
            for X, Y in some_tuples(self.basis(), 2, tester._max_runs):
                i = X.degree()
                j = Y.degree()
                Z = self.bracket(X, Y)
                if Z == 0:
                    continue
                Zdeg = Z.degree()
                tester.assertEqual(Zdeg, i + j,
                    msg="Lie bracket [%s, %s] has degree %s, not degree %s " %
                        (X, Y, Zdeg, i + j))
                tester.assertIn(
                    Z.to_vector(),
                    self.homogeneous_component_as_submodule(i + j),
                    msg="Lie bracket [%s, %s] is not in the "
                        "homogeneous component of degree %s" % (X, Y, i + j))

        @cached_method
        def homogeneous_component_as_submodule(self, d):
            r"""
            Return the ``d``-th homogeneous component of ``self``
            as a submodule.

            EXAMPLES::

                sage: C = LieAlgebras(QQ).WithBasis().Graded()
                sage: C = C.FiniteDimensional().Stratified().Nilpotent()
                sage: L = LieAlgebra(QQ, {('x','y'): {'z': 1}},
                ....:                     nilpotent=True, category=C)
                sage: L.homogeneous_component_as_submodule(2)
                Sparse vector space of degree 3 and dimension 1 over Rational Field
                Basis matrix:
                [0 0 1]
            """
            B = self.homogeneous_component_basis(d)
            return self.module().submodule([X.to_vector() for X in B])

    class Stratified(CategoryWithAxiom_over_base_ring):
        r"""
        Category of finite dimensional stratified Lie algebras with a basis.

        A stratified Lie algebra is a graded Lie algebra that is generated
        as a Lie algebra by its homogeneous component of degree 1. That is
        to say, for a graded Lie algebra `L = \bigoplus_{k=1}^M L_k`,
        we have `L_{k+1} = [L_1, L_k]`.

        EXAMPLES::

            sage: C = LieAlgebras(QQ).WithBasis().Graded().Stratified().FiniteDimensional()
            sage: C
            Category of finite dimensional stratified lie algebras with basis over Rational Field

        A finite-dimensional stratified Lie algebra is nilpotent::

            sage: C is C.Nilpotent()
            True

        TESTS::

            sage: C = LieAlgebras(QQ).WithBasis().Graded().FiniteDimensional().Stratified()
            sage: TestSuite(C).run()
        """
        class ParentMethods:
            def _test_generated_by_degree_one(self, **options):
                r"""
                Tests that the Lie algebra is generated by the homogeneous
                component of degree one.

                INPUT:

                - ``options`` -- any keyword arguments accepted by
                  :meth:`_tester`

                EXAMPLES::

                    sage: C = LieAlgebras(QQ).WithBasis().Graded()
                    sage: C = C.FiniteDimensional().Stratified().Nilpotent()
                    sage: sc = {('x','y'): {'z': 1}}
                    sage: L.<x,y,z> = LieAlgebra(QQ, sc, nilpotent=True, category=C)
                    sage: L._test_generated_by_degree_one()

                    sage: sc = {('x','y'): {'z': 1}, ('a','b'): {'c':1}, ('z','c'): {'m':1}}
                    sage: L.<a,b,c,m,x,y,z> = LieAlgebra(QQ, sc, nilpotent=True, category=C)
                    sage: L._test_generated_by_degree_one()
                    Traceback (most recent call last):
                    ...
                    AssertionError: [a, b, x, y] does not generate Nilpotent Lie algebra
                     on 7 generators (a, b, c, m, x, y, z) over Rational Field

                See the documentation for :class:`TestSuite` for more information.
                """
                tester = self._tester(**options)

                V1 = self.homogeneous_component_as_submodule(1)
                B1 = V1.basis()
                m = self.module()

                V = V1
                d = 0
                i = 0
                while V.dimension() > d:
                    if i > tester._max_runs:
                        return
                    B = V.basis()
                    d = V.dimension()
                    V = m.submodule(B + [self.bracket(X, Y).to_vector()
                                         for X in B1 for Y in B])

                tester.assertEqual(V, m,
                    msg="%s does not generate %s" % ([self(X) for X in B1], self))

            def degree_on_basis(self, m):
                r"""
                Return the degree of the basis element indexed by ``m``.

                If the degrees of the basis elements are not defined,
                they will be computed. By assumption the stratification
                `L_1 \oplus \cdots \oplus L_s` of ``self`` is such that each
                component `L_k` is spanned by some subset of the basis.

                The degree of a basis element `X` is therefore the largest
                index `k` such that `X \in L_k \oplus \cdots \oplus L_s`. The
                space  `L_k \oplus \cdots \oplus L_s` is by assumption the
                `k`-th term of the lower central series.

                EXAMPLES::

                    sage: C = LieAlgebras(QQ).WithBasis().Graded()
                    sage: C = C.FiniteDimensional().Stratified().Nilpotent()
                    sage: sc = {('X','Y'): {'Z': 1}}
                    sage: L.<X,Y,Z> = LieAlgebra(QQ, sc, nilpotent=True, category=C)
                    sage: L.degree_on_basis(X.leading_support())
                    1
                    sage: X.degree()
                    1
                    sage: Y.degree()
                    1
                    sage: L[X, Y]
                    Z
                    sage: Z.degree()
                    2
                """
                if not hasattr(self, '_basis_degrees'):
                    lcs = self.lower_central_series(submodule=True)
                    self._basis_degrees = {}

                    for k in reversed(range(len(lcs) - 1)):
                        for X in self.basis():
                            if X in self._basis_degrees:
                                continue
                            if X.to_vector() in lcs[k]:
                                self._basis_degrees[X] = k + 1

                m = self.basis()[m]
                return self._basis_degrees[m]




r"""
CW Complexes
"""
#*****************************************************************************
#  Copyright (C) 2015 Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.misc.abstract_method import abstract_method
from sage.misc.cachefunc import cached_method
from sage.categories.category_singleton import Category_singleton
from sage.categories.category_with_axiom import CategoryWithAxiom
from sage.categories.sets_cat import Sets

class CWComplexes(Category_singleton):
    r"""
    The category of CW complexes.

    A CW complex is a Closure-finite cell complex in the Weak topology.

    REFERENCES:

    - :wikipedia:`CW_complex`

    .. NOTE::

        The notion of "finite" is that the number of cells is finite.

    EXAMPLES::

        sage: from sage.categories.cw_complexes import CWComplexes
        sage: C = CWComplexes(); C
        Category of CW complexes

    TESTS::

        sage: TestSuite(C).run()
    """
    @cached_method
    def super_categories(self):
        """
        EXAMPLES::

            sage: from sage.categories.cw_complexes import CWComplexes
            sage: CWComplexes().super_categories()
            [Category of topological spaces]
        """
        return [Sets().Topological()]

    def _repr_object_names(self):
        """
        EXAMPLES::

            sage: from sage.categories.cw_complexes import CWComplexes
            sage: CWComplexes() # indirect doctest
            Category of CW complexes
        """
        return "CW complexes"

    class SubcategoryMethods:
        @cached_method
        def Connected(self):
            """
            Return the full subcategory of the connected objects of ``self``.

            EXAMPLES::

                sage: from sage.categories.cw_complexes import CWComplexes
                sage: CWComplexes().Connected()
                Category of connected CW complexes

            TESTS::

                sage: TestSuite(CWComplexes().Connected()).run()
                sage: CWComplexes().Connected.__module__
                'sage.categories.cw_complexes'
            """
            return self._with_axiom('Connected')

        @cached_method
        def FiniteDimensional(self):
            """
            Return the full subcategory of the finite dimensional
            objects of ``self``.

            EXAMPLES::

                sage: from sage.categories.cw_complexes import CWComplexes
                sage: C = CWComplexes().FiniteDimensional(); C
                Category of finite dimensional CW complexes

            TESTS::

                sage: from sage.categories.cw_complexes import CWComplexes
                sage: C = CWComplexes().FiniteDimensional()
                sage: TestSuite(C).run()
                sage: CWComplexes().Connected().FiniteDimensional.__module__
                'sage.categories.cw_complexes'
            """
            return self._with_axiom('FiniteDimensional')

    class Connected(CategoryWithAxiom):
        """
        The category of connected CW complexes.
        """

    class FiniteDimensional(CategoryWithAxiom):
        """
        Category of finite dimensional CW complexes.
        """

    class Finite(CategoryWithAxiom):
        """
        Category of finite CW complexes.

        A finite CW complex is a CW complex with a finite number of cells.
        """
        def extra_super_categories(self):
            """
            Return the extra super categories of ``self``.

            A finite CW complex is a compact finite-dimensional CW complex.

            EXAMPLES::

                sage: from sage.categories.cw_complexes import CWComplexes
                sage: C = CWComplexes().Finite()
                sage: C.extra_super_categories()
                [Category of finite dimensional CW complexes,
                 Category of compact topological spaces]
            """
            return [CWComplexes().FiniteDimensional(), Sets().Topological().Compact()]

        class ParentMethods:
            @cached_method
            def dimension(self):
                """
                Return the dimension of ``self``.

                EXAMPLES::

                    sage: from sage.categories.cw_complexes import CWComplexes
                    sage: X = CWComplexes().example()
                    sage: X.dimension()
                    2
                """
                C = self.cells()
                return max(c.dimension() for d in C.keys() for c in C[d])

    def Compact_extra_super_categories(self):
        """
        Return extraneous super categories for ``CWComplexes().Compact()``.

        A compact CW complex is finite, see Proposition A.1 in [Hat2002]_.

        .. TODO::

            Fix the name of finite CW complexes.

        EXAMPLES::

            sage: from sage.categories.cw_complexes import CWComplexes
            sage: CWComplexes().Compact() # indirect doctest
            Category of finite finite dimensional CW complexes
            sage: CWComplexes().Compact() is CWComplexes().Finite()
            True
        """
        return (Sets().Finite(),)

    class ElementMethods:
        @abstract_method
        def dimension(self):
            """
            Return the dimension of ``self``.

            EXAMPLES::

                sage: from sage.categories.cw_complexes import CWComplexes
                sage: X = CWComplexes().example()
                sage: X.an_element().dimension()
                2
            """

    class ParentMethods:
        @abstract_method
        def dimension(self):
            """
            Return the dimension of ``self``.

            EXAMPLES::

                sage: from sage.categories.cw_complexes import CWComplexes
                sage: X = CWComplexes().example()
                sage: X.dimension()
                2
            """

        @abstract_method(optional=True)
        def cells(self):
            """
            Return the cells of ``self``.

            EXAMPLES::

                sage: from sage.categories.cw_complexes import CWComplexes
                sage: X = CWComplexes().example()
                sage: C = X.cells()
                sage: sorted((d, C[d]) for d in C.keys())
                [(0, (0-cell v,)),
                 (1, (0-cell e1, 0-cell e2)),
                 (2, (2-cell f,))]
            """




r"""
Fast binary operations for basic types
"""

from .types cimport mpz_t, mpq_t
from .mpz cimport mpz_set, mpz_add, mpz_mul
from .mpq cimport mpq_canonicalize, mpq_numref, mpq_denref, mpq_add

cdef inline void mpq_add_z(mpq_t res, mpq_t op1, mpz_t op2):
    mpz_mul(mpq_numref(res), mpq_denref(op1), op2)
    mpz_add(mpq_numref(res), mpq_numref(res), mpq_numref(op1))
    mpz_set(mpq_denref(res), mpq_denref(op1))

cdef inline void mpq_div_zz(mpq_t res, mpz_t op1, mpz_t op2):
    mpz_set(mpq_numref(res), op1)
    mpz_set(mpq_denref(res), op2)
    mpq_canonicalize(res)

cdef inline void mpq_mul_z(mpq_t res, mpq_t op1, mpz_t op2):
    # (A/B) * C = (C/B) * A
    mpq_div_zz(res, op2, mpq_denref(op1))
    mpz_mul(mpq_numref(res), mpq_numref(res), mpq_numref(op1))



r"""
Finite lattice posets
"""
#*****************************************************************************
#  Copyright (C) 2011 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.categories.category_with_axiom import CategoryWithAxiom

class FiniteLatticePosets(CategoryWithAxiom):
    r"""
    The category of finite lattices, i.e. finite partially ordered
    sets which are also lattices.

    EXAMPLES::

        sage: FiniteLatticePosets()
        Category of finite lattice posets
        sage: FiniteLatticePosets().super_categories()
        [Category of lattice posets, Category of finite posets]
        sage: FiniteLatticePosets().example()
        NotImplemented

    .. SEEALSO::

        :class:`FinitePosets`, :class:`LatticePosets`, :class:`~sage.combinat.posets.lattices.FiniteLatticePoset`

    TESTS::

        sage: C = FiniteLatticePosets()
        sage: C is FiniteLatticePosets().Finite()
        True
        sage: TestSuite(C).run()

    """

    class ParentMethods:

        def join_irreducibles(self):
            r"""
            Return the join-irreducible elements of this finite lattice.

            A *join-irreducible element* of ``self`` is an element
            `x` that is not minimal and that can not be written as
            the join of two elements different from `x`.

            EXAMPLES::

                sage: L = LatticePoset({0:[1,2],1:[3],2:[3,4],3:[5],4:[5]})
                sage: L.join_irreducibles()
                [1, 2, 4]

            .. SEEALSO::

                - Dual function: :meth:`meet_irreducibles`
                - Other: :meth:`~sage.combinat.posets.lattices.FiniteLatticePoset.double_irreducibles`,
                  :meth:`join_irreducibles_poset`
            """
            return [x for x in self if len(self.lower_covers(x)) == 1]

        def join_irreducibles_poset(self):
            r"""
            Return the poset of join-irreducible elements of this finite lattice.

            A *join-irreducible element* of ``self`` is an element `x`
            that is not minimal and can not be written as the join of two
            elements different from `x`.

            EXAMPLES::

                sage: L = LatticePoset({0:[1,2,3],1:[4],2:[4],3:[4]})
                sage: L.join_irreducibles_poset()
                Finite poset containing 3 elements

            .. SEEALSO::

                - Dual function: :meth:`meet_irreducibles_poset`
                - Other: :meth:`join_irreducibles`
            """
            return self.subposet(self.join_irreducibles())

        def meet_irreducibles(self):
            r"""
            Return the meet-irreducible elements of this finite lattice.

            A *meet-irreducible element* of ``self`` is an element
            `x` that is not maximal and that can not be written as
            the meet of two elements different from `x`.

            EXAMPLES::

                sage: L = LatticePoset({0:[1,2],1:[3],2:[3,4],3:[5],4:[5]})
                sage: L.meet_irreducibles()
                [1, 3, 4]

            .. SEEALSO::

                - Dual function: :meth:`join_irreducibles`
                - Other: :meth:`~sage.combinat.posets.lattices.FiniteLatticePoset.double_irreducibles`,
                  :meth:`meet_irreducibles_poset`
            """
            return [x for x in self if len(self.upper_covers(x)) == 1]

        def meet_irreducibles_poset(self):
            r"""
            Return the poset of join-irreducible elements of this finite lattice.

            A *meet-irreducible element* of ``self`` is an element `x`
            that is not maximal and can not be written as the meet of two
            elements different from `x`.

            EXAMPLES::

                sage: L = LatticePoset({0:[1,2,3],1:[4],2:[4],3:[4]})
                sage: L.join_irreducibles_poset()
                Finite poset containing 3 elements

            .. SEEALSO::

                - Dual function: :meth:`join_irreducibles_poset`
                - Other: :meth:`meet_irreducibles`
            """
            return self.subposet(self.meet_irreducibles())

        def irreducibles_poset(self):
            """
            Return the poset of meet- or join-irreducibles of the lattice.

            A *join-irreducible* element of a lattice is an element with
            exactly one lower cover. Dually a *meet-irreducible* element
            has exactly one upper cover.

            This is the smallest poset with completion by cuts being
            isomorphic to the lattice. As a special case this returns
            one-element poset from one-element lattice.

            .. SEEALSO::

                :meth:`~sage.combinat.posets.posets.FinitePoset.completion_by_cuts`.

            EXAMPLES::

                sage: L = LatticePoset({1: [2, 3, 4], 2: [5, 6], 3: [5],
                ....:                   4: [6], 5: [9, 7], 6: [9, 8], 7: [10],
                ....:                   8: [10], 9: [10], 10: [11]})
                sage: L_ = L.irreducibles_poset()
                sage: sorted(L_)
                [2, 3, 4, 7, 8, 9, 10, 11]
                sage: L_.completion_by_cuts().is_isomorphic(L)
                True

            TESTS::

                sage: LatticePoset().irreducibles_poset()
                Finite poset containing 0 elements
                sage: posets.ChainPoset(1).irreducibles_poset()
                Finite poset containing 1 elements
            """
            if self.cardinality() == 1:
                from sage.combinat.posets.posets import Poset
                return Poset({self[0]: []})
            return self.subposet(self.join_irreducibles()+self.meet_irreducibles())

        ##########################################################################
        # Lattice morphisms

        def is_lattice_morphism(self, f, codomain):
            r"""
            Return whether ``f`` is a morphism of posets from ``self``
            to ``codomain``.

            A map `f : P \to Q` is a poset morphism if

            .. MATH::

                x \leq y \Rightarrow f(x) \leq f(y)

            for all `x,y \in P`.

            INPUT:

            - ``f`` -- a function from ``self`` to ``codomain``
            - ``codomain`` -- a lattice

            EXAMPLES:

            We build the boolean lattice of `\{2,2,3\}` and the
            lattice of divisors of `60`, and check that the map
            `b \mapsto 5 \prod_{x\in b} x` is a morphism of lattices::

                sage: D = LatticePoset((divisors(60), attrcall("divides")))
                sage: B = LatticePoset((Subsets([2,2,3]), attrcall("issubset")))
                sage: def f(b): return D(5*prod(b))
                sage: B.is_lattice_morphism(f, D)
                True

            We construct the boolean lattice `B_2`::

                sage: B = posets.BooleanLattice(2)
                sage: B.cover_relations()
                [[0, 1], [0, 2], [1, 3], [2, 3]]

            And the same lattice with new top and bottom elements
            numbered respectively `-1` and `3`::

                sage: L = LatticePoset(DiGraph({-1:[0], 0:[1,2], 1:[3], 2:[3],3:[4]}))
                sage: L.cover_relations()
                [[-1, 0], [0, 1], [0, 2], [1, 3], [2, 3], [3, 4]]

                sage: f = { B(0): L(0), B(1): L(1), B(2): L(2), B(3): L(3) }.__getitem__
                sage: B.is_lattice_morphism(f, L)
                True

                sage: f = { B(0): L(-1),B(1): L(1), B(2): L(2), B(3): L(3) }.__getitem__
                sage: B.is_lattice_morphism(f, L)
                False

                sage: f = { B(0): L(0), B(1): L(1), B(2): L(2), B(3): L(4) }.__getitem__
                sage: B.is_lattice_morphism(f, L)
                False

            .. SEEALSO::

                :meth:`~sage.categories.finite_posets.FinitePosets.ParentMethods.is_poset_morphism`
            """
            # Note: in a lattice, x <= y iff join(x,y) = y .
            # Therefore checking joins and meets is sufficient to
            # ensure that this is a poset morphism. It actually may
            # be sufficient to check just joins (or just meets).
            from sage.combinat.subset import Subsets
            for x,y in Subsets(self,2):
                if f(self.join(x,y)) != codomain.join(f(x), f(y)):
                    return False
                if f(self.meet(x,y)) != codomain.meet(f(x), f(y)):
                    return False
            return True




"""
Functors

AUTHORS:

- David Kohel and William Stein

- David Joyner (2005-12-17): examples

- Robert Bradshaw (2007-06-23): Pyrexify

- Simon King (2010-04-30): more examples, several bug fixes,
  re-implementation of the default call method,
  making functors applicable to morphisms (not only to objects)

- Simon King (2010-12): Pickling of functors without losing domain and codomain

"""

# ****************************************************************************
#  Copyright (C) 2005 David Kohel <kohel@maths.usyd.edu> and
#                     William Stein <wstein@math.ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License for more details; the full text
#  is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from . import category


def _Functor_unpickle(Cl, D, domain, codomain):
    """
    Generic unpickling function for functors.

    AUTHOR:

    - Simon King (2010-12): :trac:`10460`

    EXAMPLES::

        sage: R.<x,y> = InfinitePolynomialRing(QQ)
        sage: F = R.construction()[0]
        sage: F == loads(dumps(F))
        True
        sage: F.domain(), loads(dumps(F)).domain()
        (Category of rings, Category of rings)

    """
    F = Functor.__new__(Cl)
    Functor.__init__(F,domain,codomain)
    for s,v in D:
        setattr(F,s,v)
    return F

cdef class Functor(SageObject):
    """
    A class for functors between two categories

    NOTE:

    - In the first place, a functor is given by its domain and codomain,
      which are both categories.
    - When defining a sub-class, the user should not implement a call method.
      Instead, one should implement three methods, which are composed in the
      default call method:

      - ``_coerce_into_domain(self, x)``: Return an object of ``self``'s
        domain, corresponding to ``x``, or raise a ``TypeError``.

        - Default: Raise ``TypeError`` if ``x`` is not in ``self``'s domain.

      - ``_apply_functor(self, x)``: Apply ``self`` to an object ``x`` of
        ``self``'s domain.

        - Default: Conversion into ``self``'s codomain.

      - ``_apply_functor_to_morphism(self, f)``: Apply ``self`` to a morphism
        ``f`` in ``self``'s domain.
        - Default: Return ``self(f.domain()).hom(f,self(f.codomain()))``.

    EXAMPLES::

        sage: rings  = Rings()
        sage: abgrps = CommutativeAdditiveGroups()
        sage: F = ForgetfulFunctor(rings, abgrps)
        sage: F.domain()
        Category of rings
        sage: F.codomain()
        Category of commutative additive groups
        sage: from sage.categories.functor import is_Functor
        sage: is_Functor(F)
        True
        sage: I = IdentityFunctor(abgrps)
        sage: I
        The identity functor on Category of commutative additive groups
        sage: I.domain()
        Category of commutative additive groups
        sage: is_Functor(I)
        True

    Note that by default, an instance of the class Functor is coercion
    from the domain into the codomain. The above subclasses overloaded
    this behaviour. Here we illustrate the default::

        sage: from sage.categories.functor import Functor
        sage: F = Functor(Rings(),Fields())
        sage: F
        Functor from Category of rings to Category of fields
        sage: F(ZZ)
        Rational Field
        sage: F(GF(2))
        Finite Field of size 2

    Functors are not only about the objects of a category, but also about
    their morphisms. We illustrate it, again, with the coercion functor
    from rings to fields.

    ::

        sage: R1.<x> = ZZ[]
        sage: R2.<a,b> = QQ[]
        sage: f = R1.hom([a+b],R2)
        sage: f
        Ring morphism:
          From: Univariate Polynomial Ring in x over Integer Ring
          To:   Multivariate Polynomial Ring in a, b over Rational Field
          Defn: x |--> a + b
        sage: F(f)
        Ring morphism:
          From: Fraction Field of Univariate Polynomial Ring in x over Integer Ring
          To:   Fraction Field of Multivariate Polynomial Ring in a, b over Rational Field
          Defn: x |--> a + b
        sage: F(f)(1/x)
        1/(a + b)

    We can also apply a polynomial ring construction functor to our homomorphism. The
    result is a homomorphism that is defined on the base ring::

        sage: F = QQ['t'].construction()[0]
        sage: F
        Poly[t]
        sage: F(f)
        Ring morphism:
          From: Univariate Polynomial Ring in t over Univariate Polynomial Ring in x over Integer Ring
          To:   Univariate Polynomial Ring in t over Multivariate Polynomial Ring in a, b over Rational Field
          Defn: Induced from base ring by
                Ring morphism:
                  From: Univariate Polynomial Ring in x over Integer Ring
                  To:   Multivariate Polynomial Ring in a, b over Rational Field
                  Defn: x |--> a + b
        sage: p = R1['t']('(-x^2 + x)*t^2 + (x^2 - x)*t - 4*x^2 - x + 1')
        sage: F(f)(p)
        (-a^2 - 2*a*b - b^2 + a + b)*t^2 + (a^2 + 2*a*b + b^2 - a - b)*t - 4*a^2 - 8*a*b - 4*b^2 - a - b + 1

    """
    def __init__(self, domain, codomain):
        """
        TESTS::

            sage: from sage.categories.functor import Functor
            sage: F = Functor(Rings(),Fields())
            sage: F
            Functor from Category of rings to Category of fields
            sage: F(ZZ)
            Rational Field
            sage: F(GF(2))
            Finite Field of size 2

        """
        if not category.is_Category(domain):
            raise TypeError("domain (=%s) must be a category" % domain)
        if not category.is_Category(codomain):
            raise TypeError("codomain (=%s) must be a category" % codomain)
        self.__domain = domain
        self.__codomain = codomain

    def __hash__(self):
        r"""
        TESTS::

            sage: from sage.categories.functor import Functor
            sage: F = Functor(Rings(), Fields())
            sage: hash(F)  # random
            42
        """
        return hash(self.__domain) ^ hash(self.__codomain)

    def __reduce__(self):
        """
        Generic pickling of functors.

        AUTHOR:

        - Simon King (2010-12):  :trac:`10460`

        TESTS::

            sage: from sage.categories.pushout import CompositeConstructionFunctor
            sage: F = CompositeConstructionFunctor(QQ.construction()[0],ZZ['x'].construction()[0],QQ.construction()[0],ZZ['y'].construction()[0])
            sage: F == loads(dumps(F))
            True
            sage: F.codomain()
            Category of rings
        """
        return (_Functor_unpickle,
                (self.__class__, list(self.__dict__.items()),
                 self.__domain, self.__codomain))

    def _apply_functor(self, x):
        """
        Apply the functor to an object of ``self``'s domain.

        NOTE:

        Each subclass of :class:`Functor` should overload this method. By default,
        this method coerces into the codomain, without checking whether the
        argument belongs to the domain.

        TESTS::

            sage: from sage.categories.functor import Functor
            sage: F = Functor(FiniteFields(),Fields())
            sage: F._apply_functor(ZZ)
            Rational Field

        """
        return self.__codomain(x)

    def _apply_functor_to_morphism(self, f):
        """
        Apply the functor to a morphism between two objects of ``self``'s domain.

        NOTE:

        Each subclass of :class:`Functor` should overload this method. By
        default, this method coerces into the codomain, without checking
        whether the argument belongs to the domain.

        TESTS::

            sage: from sage.categories.functor import Functor
            sage: F = Functor(Rings(),Fields())
            sage: k.<a> = GF(25)
            sage: f = k.hom([-a-4])
            sage: R.<t> = k[]
            sage: fR = R.hom(f,R)
            sage: fF = F(fR)         # indirect doctest
            sage: fF
            Ring endomorphism of Fraction Field of Univariate Polynomial Ring in t over Finite Field in a of size 5^2
              Defn: Induced from base ring by
                    Ring endomorphism of Univariate Polynomial Ring in t over Finite Field in a of size 5^2
                      Defn: Induced from base ring by
                            Ring endomorphism of Finite Field in a of size 5^2
                              Defn: a |--> 4*a + 1
            sage: fF((a^2+a)*t^2/(a*t - a^2))
            ((4*a + 2)*t^2)/(t + a + 4)

        """
        try:
            return self(f.domain()).hom(f, self(f.codomain()))
        except Exception:
            raise TypeError('unable to transform %s into a morphism in %s' % (f,self.codomain()))

    def _coerce_into_domain(self, x):
        """
        Interprete the argument as an object of self's domain.

        NOTE:

        A subclass of :class:`Functor` may overload this method. It should
        return an object of self's domain, and should raise a ``TypeError``
        if this is impossible.

        By default, the argument will not be changed, but a ``TypeError``
        will be raised if the argument does not belong to the domain.

        TESTS::

            sage: from sage.categories.functor import Functor
            sage: F = Functor(Fields(),Fields())
            sage: F(QQ)
            Rational Field
            sage: F(ZZ) # indirect doctest
            Traceback (most recent call last):
            ...
            TypeError: x (=Integer Ring) is not in Category of fields

        """
        if not (x in  self.__domain):
            raise TypeError("x (=%s) is not in %s" % (x, self.__domain))
        return x

    def _repr_(self):
        """
        TESTS::

            sage: from sage.categories.functor import Functor
            sage: F = Functor(Rings(),Fields())
            sage: F #indirect doctest
            Functor from Category of rings to Category of fields

        A functor can be renamed if its type is a Python class
        (see :trac:`16156`)::

            sage: I = IdentityFunctor(Rings()); I
            The identity functor on Category of rings
            sage: I.rename('Id'); I
            Id

        """
        return "Functor from %s to %s"%(self.__domain, self.__codomain)

    def __call__(self, x):
        """
        NOTE:

        Implement _coerce_into_domain, _apply_functor and
        _apply_functor_to_morphism when subclassing Functor.

        TESTS:

        The default::

            sage: from sage.categories.functor import Functor
            sage: F = Functor(Rings(),Fields())
            sage: F
            Functor from Category of rings to Category of fields
            sage: F(ZZ)
            Rational Field
            sage: F(GF(2))
            Finite Field of size 2

        Two subclasses::

            sage: F1 = ForgetfulFunctor(FiniteFields(),Fields())
            sage: F1(GF(5)) #indirect doctest
            Finite Field of size 5
            sage: F1(ZZ)
            Traceback (most recent call last):
            ...
            TypeError: x (=Integer Ring) is not in Category of finite enumerated fields
            sage: F2 = IdentityFunctor(Fields())
            sage: F2(RR) is RR #indirect doctest
            True
            sage: F2(ZZ['x','y'])
            Traceback (most recent call last):
            ...
            TypeError: x (=Multivariate Polynomial Ring in x, y over Integer Ring) is not in Category of fields

        The last example shows that it is tested whether the result of
        applying the functor lies in the functor's codomain. Note that
        the matrix functor used to be defined similar to this example,
        which was fixed in :trac:`8807`::

            sage: class IllFunctor(Functor):
            ....:   def __init__(self, m,n):
            ....:       self._m = m
            ....:       self._n = n
            ....:       Functor.__init__(self,Rings(),Rings())
            ....:   def _apply_functor(self, R):
            ....:       return MatrixSpace(R,self._m,self._n)
            sage: F = IllFunctor(2,2)
            sage: F(QQ)
            Full MatrixSpace of 2 by 2 dense matrices over Rational Field
            sage: F = IllFunctor(2,3)
            sage: F(QQ)
            Traceback (most recent call last):
            ...
            TypeError: Functor from Category of rings to Category of rings is ill-defined, since it sends x (=Rational Field) to something that is not in Category of rings.

        """
        from sage.categories.morphism import is_Morphism
        if is_Morphism(x):
            return self._apply_functor_to_morphism(x)
        y = self._apply_functor(self._coerce_into_domain(x))
        if not ((y in self.__codomain) or (y in self.__codomain.Homsets())):
            raise TypeError("%s is ill-defined, since it sends x (=%s) to something that is not in %s." % (repr(self), x, self.__codomain))
        return y

    def domain(self):
        """
        The domain of self

        EXAMPLES::

            sage: F = ForgetfulFunctor(FiniteFields(),Fields())
            sage: F.domain()
            Category of finite enumerated fields

        """
        return self.__domain

    def codomain(self):
        """
        The codomain of self

        EXAMPLES::

            sage: F = ForgetfulFunctor(FiniteFields(),Fields())
            sage: F.codomain()
            Category of fields

        """
        return self.__codomain


def is_Functor(x):
    """
    Test whether the argument is a functor

    NOTE:

    There is a deprecation warning when using it from top level.
    Therefore we import it in our doc test.

    EXAMPLES::

        sage: from sage.categories.functor import is_Functor
        sage: F1 = QQ.construction()[0]
        sage: F1
        FractionField
        sage: is_Functor(F1)
        True
        sage: is_Functor(FractionField)
        False
        sage: F2 = ForgetfulFunctor(Fields(), Rings())
        sage: F2
        The forgetful functor from Category of fields to Category of rings
        sage: is_Functor(F2)
        True

    """
    return isinstance(x, Functor)


###########################################
# The natural functors in Sage
###########################################

class ForgetfulFunctor_generic(Functor):
    """
    The forgetful functor, i.e., embedding of a subcategory.

    NOTE:

    Forgetful functors should be created using :func:`ForgetfulFunctor`,
    since the init method of this class does not check whether the
    domain is a subcategory of the codomain.

    EXAMPLES::

        sage: F = ForgetfulFunctor(FiniteFields(),Fields()) #indirect doctest
        sage: F
        The forgetful functor from Category of finite enumerated fields to Category of fields
        sage: F(GF(3))
        Finite Field of size 3

    """
    def __reduce__(self):
        """
        EXAMPLES::

            sage: F = ForgetfulFunctor(Groups(), Sets())
            sage: loads(F.dumps()) == F
            True
        """
        return ForgetfulFunctor, (self.domain(), self.codomain())

    def _repr_(self):
        """
        TESTS::

            sage: F = ForgetfulFunctor(FiniteFields(),Fields())
            sage: F #indirect doctest
            The forgetful functor from Category of finite enumerated fields to Category of fields

        """
        return "The forgetful functor from %s to %s" % (self.domain(),
                                                        self.codomain())

    def __eq__(self, other):
        """
        NOTE:

        It is tested whether the second argument belongs to the class
        of forgetful functors and has the same domain and codomain as
        self. If the second argument is a functor of a different class
        but happens to be a forgetful functor, both arguments will
        still be considered as being *different*.

        TESTS::

            sage: F1 = ForgetfulFunctor(FiniteFields(),Fields())

        This is to test against a bug occurring in a previous version
        (see :trac:`8800`)::

            sage: F1 == QQ #indirect doctest
            False

        We now compare with the fraction field functor, that has a
        different domain::

            sage: F2 = QQ.construction()[0]
            sage: F1 == F2 #indirect doctest
            False
        """
        from sage.categories.pushout import IdentityConstructionFunctor
        if not isinstance(other, (self.__class__, IdentityConstructionFunctor)):
            return False
        return (self.domain() == other.domain() and
                self.codomain() == other.codomain())

    def __ne__(self, other):
        """
        Return whether ``self`` is not equal to ``other``.

        EXAMPLES::

            sage: F1 = ForgetfulFunctor(FiniteFields(),Fields())
            sage: F1 != F1
            False
            sage: F1 != QQ
            True
        """
        return not self ==  other


class IdentityFunctor_generic(ForgetfulFunctor_generic):
    """
    Generic identity functor on any category

    NOTE:

    This usually is created using :func:`IdentityFunctor`.

    EXAMPLES::

        sage: F = IdentityFunctor(Fields()) #indirect doctest
        sage: F
        The identity functor on Category of fields
        sage: F(RR) is RR
        True
        sage: F(ZZ)
        Traceback (most recent call last):
        ...
        TypeError: x (=Integer Ring) is not in Category of fields

    TESTS::

        sage: R = IdentityFunctor(Rings())
        sage: P, _ = QQ['t'].construction()
        sage: R == P
        False
        sage: P == R
        False
        sage: R == QQ
        False
    """
    def __init__(self, C):
        """
        TESTS::

            sage: from sage.categories.functor import IdentityFunctor_generic
            sage: F = IdentityFunctor_generic(Groups())
            sage: F == IdentityFunctor(Groups())
            True
            sage: F
            The identity functor on Category of groups

        """
        ForgetfulFunctor_generic.__init__(self, C, C)

    def __reduce__(self):
        """
        EXAMPLES::

            sage: F = IdentityFunctor(Groups())
            sage: loads(F.dumps()) == F
            True

        """
        return IdentityFunctor, (self.domain(), )

    def _repr_(self):
        """
        TESTS::

            sage: fields = Fields()
            sage: F = IdentityFunctor(fields)
            sage: F #indirect doctest
            The identity functor on Category of fields

        """
        return "The identity functor on %s"%(self.domain())

    def _apply_functor(self, x):
        """
        Apply the functor to an object of ``self``'s domain.

        TESTS::

            sage: fields = Fields()
            sage: F = IdentityFunctor(fields)
            sage: F._apply_functor(QQ)
            Rational Field

        It is not tested here whether the argument belongs to the domain
        (this test is done in the default method ``_coerce_into_domain``)::

            sage: F._apply_functor(ZZ)
            Integer Ring

        """
        return x

def IdentityFunctor(C):
    """
    Construct the identity functor of the given category.

    INPUT:

    A category, ``C``.

    OUTPUT:

    The identity functor in ``C``.

    EXAMPLES::

        sage: rings = Rings()
        sage: F = IdentityFunctor(rings)
        sage: F(ZZ['x','y']) is ZZ['x','y']
        True

    """
    return IdentityFunctor_generic(C)

def ForgetfulFunctor(domain, codomain):
    """
    Construct the forgetful function from one category to another.

    INPUT:

    ``C``, ``D`` - two categories

    OUTPUT:

    A functor that returns the corresponding object of ``D`` for
    any element of ``C``, by forgetting the extra structure.

    ASSUMPTION:

    The category ``C`` must be a sub-category of ``D``.

    EXAMPLES::

        sage: rings = Rings()
        sage: abgrps = CommutativeAdditiveGroups()
        sage: F = ForgetfulFunctor(rings, abgrps)
        sage: F
        The forgetful functor from Category of rings to Category of commutative additive groups

    It would be a mistake to call it in opposite order::

        sage: F = ForgetfulFunctor(abgrps, rings)
        Traceback (most recent call last):
        ...
        ValueError: Forgetful functor not supported for domain Category of commutative additive groups

    If both categories are equal, the forgetful functor is the same as the
    identity functor::

        sage: ForgetfulFunctor(abgrps, abgrps) == IdentityFunctor(abgrps)
        True

    """
    if domain == codomain:
        return IdentityFunctor(domain)
    if not domain.is_subcategory(codomain):
        raise ValueError("Forgetful functor not supported for domain %s" % domain)
    return ForgetfulFunctor_generic(domain, codomain)




r"""
Cython-like rich comparisons in Python

With "rich comparisons", we mean the Python 3 comparisons which are
usually implemented in Python using methods like ``__eq__`` and
``__lt__``. Internally in Python, there is only one rich comparison
slot ``tp_richcompare``. The actual operator is passed as an integer
constant (defined in this module as
``op_LT``, ``op_LE``, ``op_EQ``, ``op_NE``, ``op_GT``, ``op_GE``).

Cython exposes rich comparisons in ``cdef`` classes as the
``__richcmp__`` special method. The Sage coercion model also supports
rich comparisons this way: for two instances ``x`` and ``y``
of :class:`~sage.structure.element.Element`, ``x._richcmp_(y, op)``
is called when the user does something like ``x <= y``
(possibly after coercion if ``x`` and ``y`` have different parents).

Various helper functions exist to make it easier to implement rich
comparison: the most important one is the :func:`richcmp` function.
This is analogous to the Python 2 function ``cmp()`` but implements
rich comparison, with the comparison operator (e.g. ``op_GE``) as
third argument. There is also :func:`richcmp_not_equal` which is like
:func:`richcmp` but it is optimized assuming that the compared objects
are not equal.

The functions :func:`rich_to_bool` and :func:`rich_to_bool_sgn` can be
used to convert results of ``cmp()`` (i.e. -1, 0 or 1) to a boolean
``True``/``False`` for rich comparisons.

AUTHORS:

- Jeroen Demeyer
"""

#*****************************************************************************
#       Copyright (C) 2017-2018 Jeroen Demeyer <J.Demeyer@UGent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from cpython.object cimport Py_TYPE, PyTypeObject
from sage.cpython.wrapperdescr cimport get_slotdef, wrapperbase, PyDescr_NewWrapper

cdef extern from *:
    void PyType_Modified(PyTypeObject* cls)


op_LT = Py_LT   # operator <
op_LE = Py_LE   # operator <=
op_EQ = Py_EQ   # operator ==
op_NE = Py_NE   # operator !=
op_GT = Py_GT   # operator >
op_GE = Py_GE   # operator >=


# Slotdefs for richcmp methods (the "bytes" below is arbitrary,
# any type which implements these methods would work)
cdef wrapperbase* richcmp_slotdef[6]
richcmp_slotdef[Py_EQ] = get_slotdef(bytes.__eq__)
richcmp_slotdef[Py_NE] = get_slotdef(bytes.__ne__)
richcmp_slotdef[Py_LT] = get_slotdef(bytes.__lt__)
richcmp_slotdef[Py_GT] = get_slotdef(bytes.__gt__)
richcmp_slotdef[Py_LE] = get_slotdef(bytes.__le__)
richcmp_slotdef[Py_GE] = get_slotdef(bytes.__ge__)


cpdef richcmp_item(x, y, int op):
    """
    This function is meant to implement lexicographic rich comparison
    of sequences (lists, vectors, polynomials, ...).
    The inputs ``x`` and ``y`` are corresponding items of such lists
    which should compared.

    INPUT:

    - ``x``, ``y`` -- arbitrary Python objects. Typically, these are
      ``X[i]`` and ``Y[i]`` for sequences ``X`` and ``Y``.

    - ``op`` -- comparison operator (one of ``op_LT`, ``op_LE``,
      ``op_EQ``, ``op_NE``, ``op_GT``, ``op_GE``)

    OUTPUT:

    Assuming that ``x = X[i]`` and ``y = Y[i]``:

    - if the comparison ``X {op} Y`` (where ``op`` is the given
      operation) could not be decided yet (i.e. we should compare the
      next items in the list): return ``NotImplemented``

    - otherwise, if the comparison ``X {op} Y`` could be decided:
      return ``x {op} y``, which should then also be the result for
      ``X {op} Y``.

    .. NOTE::

        Since ``x {op} y`` cannot return ``NotImplemented``, the two
        cases above are mutually exclusive.

    The semantics of the comparison is different from Python lists or
    tuples in the case that the order is not total. Assume that ``A``
    and ``B`` are lists whose rich comparison is implemented using
    ``richcmp_item`` (as in EXAMPLES below). Then

    - ``A == B`` iff ``A[i] == B[i]`` for all indices `i`.

    - ``A != B`` iff ``A[i] != B[i]`` for some index `i`.

    - ``A < B`` iff ``A[i] < B[i]`` for some index `i` and
      for all `j < i`, ``A[j] <= B[j]``.

    - ``A <= B`` iff ``A < B`` or ``A[i] <= B[i]`` for all `i`.

    - ``A > B`` iff ``A[i] > B[i]`` for some index `i` and
      for all `j < i`, ``A[j] >= B[j]``.

    - ``A >= B`` iff ``A > B`` or ``A[i] >= B[i]`` for all `i`.

    See below for a detailed description of the exact semantics of
    ``richcmp_item`` in general.

    EXAMPLES::

        sage: from sage.structure.richcmp import *
        sage: @richcmp_method
        ....: class Listcmp(list):
        ....:     def __richcmp__(self, other, op):
        ....:         for i in range(len(self)):  # Assume equal lengths
        ....:             res = richcmp_item(self[i], other[i], op)
        ....:             if res is not NotImplemented:
        ....:                 return res
        ....:         return rich_to_bool(op, 0)  # Consider the lists to be equal
        sage: a = Listcmp([0, 1, 3])
        sage: b = Listcmp([0, 2, 1])
        sage: a == a
        True
        sage: a != a
        False
        sage: a < a
        False
        sage: a <= a
        True
        sage: a > a
        False
        sage: a >= a
        True
        sage: a == b, b == a
        (False, False)
        sage: a != b, b != a
        (True, True)
        sage: a < b, b > a
        (True, True)
        sage: a <= b, b >= a
        (True, True)
        sage: a > b, b < a
        (False, False)
        sage: a >= b, b <= a
        (False, False)

    The above tests used a list of integers, where the result of
    comparisons are the same as for Python lists.

    If we want to see the difference, we need more general entries in
    the list. The comparison rules are made to be consistent with
    setwise operations. If `A` and `B` are sets, we define ``A {op} B``
    to be true if ``a {op} B`` is true for every `a` in `A` and
    `b` in `B`. Interval comparisons are a special case of this. For
    lists of non-empty(!) sets, we want that ``[A1, A2] {op} [B1, B2]``
    is true if and only if ``[a1, a2] {op} [b1, b2]`` is true for all
    elements. We verify this::

        sage: @richcmp_method
        ....: class Setcmp(tuple):
        ....:     def __richcmp__(self, other, op):
        ....:         return all(richcmp(x, y, op) for x in self for y in other)
        sage: sym = {op_EQ: "==", op_NE: "!=", op_LT: "<", op_GT: ">", op_LE: "<=", op_GE: ">="}
        sage: for A1 in Set(range(4)).subsets():  # long time
        ....:     if not A1: continue
        ....:     for B1 in Set(range(4)).subsets():
        ....:         if not B1: continue
        ....:         for A2 in Set(range(4)).subsets():
        ....:             if not A2: continue
        ....:             for B2 in Set(range(3)).subsets():
        ....:                 if not B2: continue
        ....:                 L1 = Listcmp([Setcmp(A1), Setcmp(A2)])
        ....:                 L2 = Listcmp([Setcmp(B1), Setcmp(B2)])
        ....:                 for op in range(6):
        ....:                     reslist = richcmp(L1, L2, op)
        ....:                     reselt = all(richcmp([a1, a2], [b1, b2], op) for a1 in A1 for a2 in A2 for b1 in B1 for b2 in B2)
        ....:                     assert reslist is reselt

    EXACT SEMANTICS:

    Above, we only described how ``richcmp_item`` behaves when it is
    used to compare sequences. Here, we specify the exact semantics.
    First of all, recall that the result of ``richcmp_item(x, y, op)``
    is either ``NotImplemented`` or ``x {op} y``.

    - if ``op`` is ``==``: return ``NotImplemented`` if ``x == y``.
      If ``x == y`` is false, then return ``x == y``.

    - if ``op`` is ``!=``: return ``NotImplemented`` if not ``x != y``.
      If ``x != y`` is true, then return ``x != y``.

    - if ``op`` is ``<``: return ``NotImplemented`` if ``x == y``.
      If ``x < y`` or not ``x <= y``, return ``x < y``.
      Otherwise (if both ``x == y`` and ``x < y`` are false but
      ``x <= y`` is true), return ``NotImplemented``.

    - if ``op`` is ``<=``: return ``NotImplemented`` if ``x == y``.
      If ``x < y`` or not ``x <= y``, return ``x <= y``.
      Otherwise (if both ``x == y`` and ``x < y`` are false but
      ``x <= y`` is true), return ``NotImplemented``.

    - the ``>`` and ``>=`` operators are analogous to ``<`` and ``<=``.
    """
    if op == Py_NE:
        res = (x != y)
        if not res:
            return NotImplemented
        return res  # (x != y)  --> True

    # If x and y are equal, we cannot decide
    res = (x == y)
    if res:
        return NotImplemented

    if op == Py_EQ:
        return res  # not (x == y)  --> False

    # At this point, {op} is < or <= or > or >=. In the comments below,
    # we always refer to < and <= but > and >= are obviously analogous.

    # Compute x {op} y and convert to boolean (0 or 1)
    res = PyObject_RichCompare(x, y, op)
    cdef bint bres = res

    # true (1) for <= and >= and false (0) for < and >
    cdef bint op_is_not_strict = op & 1

    if bres != op_is_not_strict:
        # If we are asked to compute (x < y) and (x < y) is true,
        # return (x < y) which is true.
        # If we are asked to compute (x <= y) and (x <= y) is false,
        # return (x <= y) which is false.
        return res

    # Finally, check the inequality with the other strictness
    # (< becomes <= and vice versa; this corresponds to replacing op by
    # op ^ 1). This check is redundant in the typical case that (x <= y)
    # is equivalent to (x < y or x == y): since (x == y) returned false,
    # we expect that this PyObject_RichCompare() call returns the same
    # boolean result as the previous one.
    cdef bint xres = PyObject_RichCompare(x, y, op ^ 1)
    if xres == bres:  # As expected
        return res

    # OK, we are in a special case now. We checked that (x == y) is
    # false, that (x < y) is false but (x <= y) is true.
    # Since we want to give more importance to the < and <= results than
    # the == result, we treat this case as equality. Therefore, we
    # cannot decide.
    return NotImplemented


cdef slot_tp_richcompare(self, other, int op):
    """
    Function to put in the ``tp_richcompare`` slot.
    """
    return self.__richcmp__(other, op)


def richcmp_method(cls):
    """
    Class decorator to implement rich comparison using the special
    method ``__richcmp__`` (analogous to Cython) instead of the 6
    methods ``__eq__`` and friends.

    This changes the class in-place and returns the given class.

    EXAMPLES::

        sage: from sage.structure.richcmp import *
        sage: sym = {op_EQ: "==", op_NE: "!=", op_LT: "<", op_GT: ">", op_LE: "<=", op_GE: ">="}
        sage: @richcmp_method
        ....: class A(str):
        ....:     def __richcmp__(self, other, op):
        ....:         print("%s %s %s" % (self, sym[op], other))
        sage: A("left") < A("right")
        left < right
        sage: object() <= A("right")
        right >= <object object at ...>

    We can call this comparison with the usual Python special methods::

        sage: x = A("left"); y = A("right")
        sage: x.__eq__(y)
        left == right
        sage: A.__eq__(x, y)
        left == right

    Everything still works in subclasses::

        sage: class B(A):
        ....:     pass
        sage: x = B("left"); y = B("right")
        sage: x != y
        left != right
        sage: x.__ne__(y)
        left != right
        sage: B.__ne__(x, y)
        left != right

    We can override ``__richcmp__`` with standard Python rich
    comparison methods and conversely::

        sage: class C(A):
        ....:     def __ne__(self, other):
        ....:         return False
        sage: C("left") != C("right")
        False
        sage: C("left") == C("right")  # Calls __eq__ from class A
        left == right

        sage: class Base(object):
        ....:     def __eq__(self, other):
        ....:         return False
        sage: @richcmp_method
        ....: class Derived(Base):
        ....:     def __richcmp__(self, other, op):
        ....:         return True
        sage: Derived() == Derived()
        True

    TESTS::

        sage: richcmp_method(None)
        Traceback (most recent call last):
        ...
        TypeError: None is not a class

        sage: @richcmp_method
        ....: class X(object):
        ....:     def __eq__(self, other):
        ....:         pass
        ....:     def __richcmp__(self, other, op):
        ....:         pass
        Traceback (most recent call last):
        ...
        TypeError: class <class '__main__.X'> defines __eq__ which cannot be combined with @richcmp_method
    """
    if not isinstance(cls, type):
        raise TypeError(f"{cls!r} is not a class")
    cdef PyTypeObject* tp = <PyTypeObject*>cls
    tp.tp_richcompare = slot_tp_richcompare

    # Install slot wrappers in the class dict
    cdef dict D = <dict>(tp.tp_dict)
    cdef wrapperbase* slotdef
    for slotdef in richcmp_slotdef:
        name = <object>slotdef.name_strobj
        if name in D:
            raise TypeError("class %r defines %s which cannot be combined with @richcmp_method" % (cls, name))
        D[name] = PyDescr_NewWrapper(tp, slotdef, <void*>slot_tp_richcompare)

    PyType_Modified(tp)

    return cls


def richcmp_by_eq_and_lt(eq_attr, lt_attr):
    r"""
    Create a rich comparison method for a partial order, where the
    order is specified by methods called ``eq_attr`` and ``lt_attr``.

    INPUT when creating the method:

    - ``eq_attr`` -- attribute name for equality comparison

    - ``lt_attr`` -- attribute name for less-than comparison

    INPUT when calling the method:

    - ``self`` -- objects having methods ``eq_attr`` and ``lt_attr``

    - ``other`` -- arbitrary object. If it does have ``eq_attr`` and
      ``lt_attr`` methods, these are used for the comparison. Otherwise,
      the comparison is undefined.

    - ``op`` -- a rich comparison operation (e.g. ``op_EQ``)

    .. NOTE::

        For efficiency, identical objects (when ``self is other``)
        always compare equal.

    .. NOTE::

        The order is partial, so ``x <= y`` is implemented as
        ``x == y or x < y``. It is not required that this is the
        negation of ``y < x``.

    .. NOTE::

        This function is intended to be used as a method ``_richcmp_``
        in a class derived from :class:`sage.structure.element.Element`
        or a method ``__richcmp__`` in a class using
        :func:`richcmp_method`.

    EXAMPLES::

        sage: from sage.structure.richcmp import richcmp_by_eq_and_lt
        sage: from sage.structure.element import Element

        sage: class C(Element):
        ....:     def __init__(self, a, b):
        ....:         super(C, self).__init__(ZZ)
        ....:         self.a = a
        ....:         self.b = b
        ....:     _richcmp_ = richcmp_by_eq_and_lt("eq", "lt")
        ....:     def eq(self, other):
        ....:         return self.a == other.a and self.b == other.b
        ....:     def lt(self, other):
        ....:         return self.a < other.a and self.b < other.b

        sage: x = C(1,2); y = C(2,1); z = C(3,3)

        sage: x == x, x <= x, x == C(1,2), x <= C(1,2)  # indirect doctest
        (True, True, True, True)
        sage: y == z, y != z
        (False, True)

        sage: x < y, y < x, x > y, y > x, x <= y, y <= x, x >= y, y >= x
        (False, False, False, False, False, False, False, False)
        sage: y < z, z < y, y > z, z > y, y <= z, z <= y, y >= z, z >= y
        (True, False, False, True, True, False, False, True)
        sage: z < x, x < z, z > x, x > z, z <= x, x <= z, z >= x, x >= z
        (False, True, True, False, False, True, True, False)

    A simple example using ``richcmp_method``::

        sage: from sage.structure.richcmp import richcmp_method, richcmp_by_eq_and_lt
        sage: @richcmp_method
        ....: class C(object):
        ....:     __richcmp__ = richcmp_by_eq_and_lt("_eq", "_lt")
        ....:     def _eq(self, other):
        ....:         return True
        ....:     def _lt(self, other):
        ....:         return True
        sage: a = C(); b = C()
        sage: a == b
        True
        sage: a > b  # Calls b._lt(a)
        True
        sage: class X(object): pass
        sage: x = X()
        sage: a == x  # Does not call a._eq(x) because x does not have _eq
        False
    """
    def richcmp(self, other, int op):
        if self is other:
            return rich_to_bool(op, 0)

        cdef bint equal_types = (type(self) is type(other))
        # Check whether "other" is of the right type
        if not equal_types:
            try:
                other_eq = getattr(other, eq_attr)
                other_lt = getattr(other, lt_attr)
            except AttributeError:
                return NotImplemented

        # Check equality first if needed
        if op != Py_LT and op != Py_GT:
            if equal_types:
                other_eq = getattr(other, eq_attr)
            if other_eq(self):
                return rich_to_bool(op, 0)
            if op == Py_EQ:
                return False
            if op == Py_NE:
                return True

        if op == Py_LT or op == Py_LE:
            self_lt = getattr(self, lt_attr)
            return self_lt(other)
        else:
            if equal_types:
                other_lt = getattr(other, lt_attr)
            return other_lt(self)

    return richcmp



"""
Repr formatting support
"""


def coeff_repr(c, is_latex=False):
    r"""
    String representing coefficients in a linear combination.

    INPUT:

    - ``c`` -- a coefficient (i.e., an element of a ring)

    OUTPUT:

    A string

    EXAMPLES::

        sage: from sage.misc.repr import coeff_repr
        sage: coeff_repr(QQ(1/2))
        '1/2'
        sage: coeff_repr(-x^2)
        '(-x^2)'
        sage: coeff_repr(QQ(1/2), is_latex=True)
        '\\frac{1}{2}'
        sage: coeff_repr(-x^2, is_latex=True)
        '\\left(-x^{2}\\right)'
    """
    if not is_latex:
        try:
            return c._coeff_repr()
        except AttributeError:
            pass
    if isinstance(c, (int, float)):
        return str(c)
    if is_latex and hasattr(c, '_latex_'):
        s = c._latex_()
    else:
        s = str(c).replace(' ', '')
    if s.find("+") != -1 or s.find("-") != -1:
        if is_latex:
            return "\\left(%s\\right)" % s
        else:
            return "(%s)" % s
    return s


def repr_lincomb(terms, is_latex=False, scalar_mult="*", strip_one=False,
                 repr_monomial=None, latex_scalar_mult=None):
    """
    Compute a string representation of a linear combination of some
    formal symbols.

    INPUT:

    - ``terms`` -- list of terms, as pairs (support, coefficient)
    - ``is_latex`` -- whether to produce latex (default: ``False``)
    - ``scalar_mult`` -- string representing the multiplication (default:``'*'``)
    - ``latex_scalar_mult`` -- latex string representing the multiplication
      (default: a space if ``scalar_mult`` is ``'*'``; otherwise ``scalar_mult``)
    - ``coeffs`` -- for backward compatibility

    OUTPUT:

    -  ``str`` - a string

    EXAMPLES::

        sage: repr_lincomb([('a',1), ('b',-2), ('c',3)])
        'a - 2*b + 3*c'
        sage: repr_lincomb([('a',0), ('b',-2), ('c',3)])
        '-2*b + 3*c'
        sage: repr_lincomb([('a',0), ('b',2), ('c',3)])
        '2*b + 3*c'
        sage: repr_lincomb([('a',1), ('b',0), ('c',3)])
        'a + 3*c'
        sage: repr_lincomb([('a',-1), ('b','2+3*x'), ('c',3)])
        '-a + (2+3*x)*b + 3*c'
        sage: repr_lincomb([('a', '1+x^2'), ('b', '2+3*x'), ('c', 3)])
        '(1+x^2)*a + (2+3*x)*b + 3*c'
        sage: repr_lincomb([('a', '1+x^2'), ('b', '-2+3*x'), ('c', 3)])
        '(1+x^2)*a + (-2+3*x)*b + 3*c'
        sage: repr_lincomb([('a', 1), ('b', -2), ('c', -3)])
        'a - 2*b - 3*c'
        sage: t = PolynomialRing(RationalField(),'t').gen()
        sage: repr_lincomb([('a', -t), ('s', t - 2), ('', t^2 + 2)])
        '-t*a + (t-2)*s + (t^2+2)'

    Examples for ``scalar_mult``::

        sage: repr_lincomb([('a',1), ('b',2), ('c',3)], scalar_mult='*')
        'a + 2*b + 3*c'
        sage: repr_lincomb([('a',2), ('b',0), ('c',-3)], scalar_mult='**')
        '2**a - 3**c'
        sage: repr_lincomb([('a',-1), ('b',2), ('c',3)], scalar_mult='**')
        '-a + 2**b + 3**c'

    Examples for ``scalar_mult`` and ``is_latex``::

        sage: repr_lincomb([('a',-1), ('b',2), ('c',3)], is_latex=True)
        '-a + 2 b + 3 c'
        sage: repr_lincomb([('a',-1), ('b',-1), ('c',3)], is_latex=True, scalar_mult='*')
        '-a - b + 3 c'
        sage: repr_lincomb([('a',-1), ('b',2), ('c',-3)], is_latex=True, scalar_mult='**')
        '-a + 2**b - 3**c'
        sage: repr_lincomb([('a',-2), ('b',-1), ('c',-3)], is_latex=True, latex_scalar_mult='*')
        '-2*a - b - 3*c'
        sage: repr_lincomb([('a',-2), ('b',-1), ('c',-3)], is_latex=True, latex_scalar_mult='')
        '-2a - b - 3c'

    Examples for ``strip_one``::

        sage: repr_lincomb([ ('a',1), (1,-2), ('3',3) ])
        'a - 2*1 + 3*3'
        sage: repr_lincomb([ ('a',-1), (1,1), ('3',3) ])
        '-a + 1 + 3*3'
        sage: repr_lincomb([ ('a',1), (1,-2), ('3',3) ], strip_one = True)
        'a - 2 + 3*3'
        sage: repr_lincomb([ ('a',-1), (1,1), ('3',3) ], strip_one = True)
        '-a + 1 + 3*3'
        sage: repr_lincomb([ ('a',1), (1,-1), ('3',3) ], strip_one = True)
        'a - 1 + 3*3'

    Examples for ``repr_monomial``::

        sage: repr_lincomb([('a',1), ('b',2), ('c',3)], repr_monomial = lambda s: s+"1")
        'a1 + 2*b1 + 3*c1'

    TESTS:

    Verify that :trac:`31672` is fixed::

        sage: alpha = var("alpha")
        sage: repr_lincomb([(x, alpha)], is_latex=True)
        '\\alpha x'
        sage: A.<psi> = PolynomialRing(QQ)
        sage: B.<t> = FreeAlgebra(A)
        sage: (psi * t)._latex_()
        '\\psi t'
    """
    # Setting scalar_mult: symbol used for scalar multiplication
    if is_latex:
        if latex_scalar_mult is not None:
            scalar_mult = latex_scalar_mult
        elif scalar_mult == "*":
            scalar_mult = " "

    if repr_monomial is None:
        if is_latex:

            def repr_monomial(monomial):
                return monomial._latex_() if hasattr(monomial, '_latex_') else str(monomial)
        else:
            repr_monomial = str

    s = ""
    first = True

    if scalar_mult is None:
        scalar_mult = "" if is_latex else "*"

    for (monomial, c) in terms:
        if c != 0:
            coeff = coeff_repr(c)
            negative = False
            if len(coeff) and coeff[0] == "-":
                negative = True
            try:
                if c < 0:
                    negative = True
            except (NotImplementedError, TypeError):
                # comparisons may not be implemented for some coefficients
                pass
            if negative:
                coeff = coeff_repr(-c, is_latex)
            else:
                coeff = coeff_repr(c, is_latex)
            if coeff == "1":
                coeff = ""
            if coeff != "0":
                if negative:
                    if first:
                        sign = "-"  # add trailing space?
                    else:
                        sign = " - "
                else:
                    if first:
                        sign = ""
                    else:
                        sign = " + "
                b = repr_monomial(monomial)
                if len(b):
                    if coeff != "":
                        if b == "1" and strip_one:
                            b = ""
                        else:
                            b = scalar_mult + b
                s += "%s%s%s" % (sign, coeff, b)
                first = False
    if first:
        return "0"
        # this can happen only if are only terms with coeff_repr(c) == "0"
    # elif s == "":
        # return "1"  # is empty string representation invalid?
    else:
        return s



r"""
Number fields
"""
# ****************************************************************************
#  Copyright (C) 2005      David Kohel <kohel@maths.usyd.edu>
#                          William Stein <wstein@math.ucsd.edu>
#                2008      Teresa Gomez-Diaz (CNRS) <Teresa.Gomez-Diaz@univ-mlv.fr>
#                2008-2009 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.categories.category_singleton import Category_singleton
from sage.categories.basic import Fields


class NumberFields(Category_singleton):
    r"""
    The category of number fields.

    EXAMPLES:

    We create the category of number fields::

        sage: C = NumberFields()
        sage: C
        Category of number fields

    By definition, it is infinite::

        sage: NumberFields().Infinite() is NumberFields()
        True

    Notice that the rational numbers `\QQ` *are* considered as
    an object in this category::

        sage: RationalField() in C
        True

    However, we can define a degree 1 extension of `\QQ`, which is of
    course also in this category::

        sage: x = PolynomialRing(RationalField(), 'x').gen()
        sage: K = NumberField(x - 1, 'a'); K
        Number Field in a with defining polynomial x - 1
        sage: K in C
        True

    Number fields all lie in this category, regardless of the name
    of the variable::

        sage: K = NumberField(x^2 + 1, 'a')
        sage: K in C
        True

    TESTS::

        sage: TestSuite(NumberFields()).run()
    """

    def super_categories(self):
        """
        EXAMPLES::

            sage: NumberFields().super_categories()
            [Category of infinite fields]
        """
        return [Fields().Infinite()]

    def __contains__(self, x):
        r"""
        Returns True if ``x`` is a number field.

        EXAMPLES::

            sage: NumberField(x^2+1,'a') in NumberFields()
            True
            sage: QuadraticField(-97,'theta') in NumberFields()
            True
            sage: CyclotomicField(97) in NumberFields()
            True

        Note that the rational numbers QQ are a number field::

            sage: QQ in NumberFields()
            True
            sage: ZZ in NumberFields()
            False
        """
        import sage.rings.number_field.number_field_base
        return sage.rings.number_field.number_field_base.is_NumberField(x)

    def _call_(self, x):
        r"""
        Constructs an object in this category from the data in ``x``,
        or throws a TypeError.

        EXAMPLES::

            sage: C = NumberFields()

            sage: C(QQ)
            Rational Field

            sage: C(NumberField(x^2+1,'a'))
            Number Field in a with defining polynomial x^2 + 1

            sage: C(UnitGroup(NumberField(x^2+1,'a')))  # indirect doctest
            Number Field in a with defining polynomial x^2 + 1

            sage: C(ZZ)
            Traceback (most recent call last):
            ...
            TypeError: unable to canonically associate a number field to Integer Ring
        """
        try:
            return x.number_field()
        except AttributeError:
            raise TypeError("unable to canonically associate a number field to %s" % x)

    class ParentMethods:
        def zeta_function(self, prec=53,
                          max_imaginary_part=0,
                          max_asymp_coeffs=40, algorithm='pari'):
            r"""
            Return the Dedekind zeta function of this number field.

            Actually, this returns an interface for computing with the
            Dedekind zeta function `\zeta_F(s)` of the number field `F`.

            INPUT:

            - ``prec`` -- optional integer (default 53) bits precision

            - ``max_imaginary_part`` -- optional real number (default 0)

            - ``max_asymp_coeffs`` -- optional integer (default 40)

            - ``algorithm`` -- optional (default "pari") either "gp" or "pari"

            OUTPUT: The zeta function of this number field.

            If algorithm is "gp", this returns an interface to Tim
            Dokchitser's gp script for computing with L-functions.

            If algorithm is "pari", this returns instead an interface to Pari's
            own general implementation of L-functions.

            EXAMPLES::

                sage: K.<a> = NumberField(ZZ['x'].0^2+ZZ['x'].0-1)
                sage: Z = K.zeta_function(); Z
                PARI zeta function associated to Number Field in a with defining polynomial x^2 + x - 1
                sage: Z(-1)
                0.0333333333333333
                sage: L.<a, b, c> = NumberField([x^2 - 5, x^2 + 3, x^2 + 1])
                sage: Z = L.zeta_function()
                sage: Z(5)
                1.00199015670185

            Using the algorithm "pari"::

                sage: K.<a> = NumberField(ZZ['x'].0^2+ZZ['x'].0-1)
                sage: Z = K.zeta_function(algorithm="pari")
                sage: Z(-1)
                0.0333333333333333
                sage: L.<a, b, c> = NumberField([x^2 - 5, x^2 + 3, x^2 + 1])
                sage: Z = L.zeta_function(algorithm="pari")
                sage: Z(5)
                1.00199015670185

            TESTS::

                sage: QQ.zeta_function()
                PARI zeta function associated to Rational Field
            """
            if algorithm == 'gp':
                from sage.lfunctions.all import Dokchitser
                r1, r2 = self.signature()
                zero = [0]
                one = [1]
                Z = Dokchitser(conductor=abs(self.absolute_discriminant()),
                               gammaV=(r1 + r2) * zero + r2 * one,
                               weight=1,
                               eps=1,
                               poles=[1],
                               prec=prec)
                s = 'nf = nfinit(%s);' % self.absolute_polynomial()
                s += 'dzk = dirzetak(nf,cflength());'
                Z.init_coeffs('dzk[k]', pari_precode=s,
                              max_imaginary_part=max_imaginary_part,
                              max_asymp_coeffs=max_asymp_coeffs)
                Z.check_functional_equation()
                Z.rename('Dokchitser Zeta function associated to %s' % self)
                return Z

            if algorithm == 'pari':
                from sage.lfunctions.pari import lfun_number_field, LFunction
                Z = LFunction(lfun_number_field(self), prec=prec)
                Z.rename('PARI zeta function associated to %s' % self)
                return Z

            raise ValueError('algorithm must be "gp" or "pari"')

    class ElementMethods:
        pass



r"""
Quantum Group Representations

AUTHORS:

- Travis Scrimshaw (2018): initial version
"""

#*****************************************************************************
#       Copyright (C) 2018 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************


from sage.misc.abstract_method import abstract_method
from sage.misc.cachefunc import cached_method
from sage.categories.modules import Modules
from sage.categories.category_types import Category_module
from sage.categories.category_with_axiom import CategoryWithAxiom_over_base_ring
from sage.categories.tensor import tensor, TensorProductsCategory

class QuantumGroupRepresentations(Category_module):
    """
    The category of quantum group representations.
    """
    @cached_method
    def super_categories(self):
        """
        Return the super categories of ``self``.

        EXAMPLES::

            sage: from sage.categories.quantum_group_representations import QuantumGroupRepresentations
            sage: QuantumGroupRepresentations(ZZ['q'].fraction_field()).super_categories()
            [Category of vector spaces over
             Fraction Field of Univariate Polynomial Ring in q over Integer Ring]
        """
        return [Modules(self.base_ring())]

    def example(self):
        """
        Return an example of a quantum group representation as per
        :meth:`Category.example <sage.categories.category.Category.example>`.

        EXAMPLES::

            sage: from sage.categories.quantum_group_representations import QuantumGroupRepresentations
            sage: Cat = QuantumGroupRepresentations(ZZ['q'].fraction_field())
            sage: Cat.example()
            V((2, 1, 0))
        """
        from sage.algebras.quantum_groups.representations import AdjointRepresentation
        from sage.combinat.crystals.tensor_product import CrystalOfTableaux
        T = CrystalOfTableaux(['A',2], shape=[2,1])
        return AdjointRepresentation(self.base_ring(), T)

    class WithBasis(CategoryWithAxiom_over_base_ring):
        """
        The category of quantum group representations with a
        distinguished basis.
        """
        class TensorProducts(TensorProductsCategory):
            """
            The category of quantum group representations with a
            distinguished basis constructed by tensor product of
            quantum group representations with a distinguished basis.
            """
            @cached_method
            def extra_super_categories(self):
                """
                EXAMPLES::

                    sage: from sage.categories.quantum_group_representations import QuantumGroupRepresentations
                    sage: Cat = QuantumGroupRepresentations(ZZ['q'].fraction_field())
                    sage: Cat.WithBasis().TensorProducts().extra_super_categories()
                    [Category of quantum group representations with basis over
                     Fraction Field of Univariate Polynomial Ring in q over Integer Ring]
                """
                return [self.base_category()]

            class ParentMethods:
                def e_on_basis(self, i, b):
                    r"""
                    Return the action of `e_i` on the basis element
                    indexed by ``b``.

                    INPUT:

                    - ``i`` -- an element of the index set
                    - ``b`` -- an element of basis keys

                    EXAMPLES::

                        sage: from sage.algebras.quantum_groups.representations import \
                        ....:  MinusculeRepresentation, AdjointRepresentation
                        sage: R = ZZ['q'].fraction_field()
                        sage: CM = crystals.Tableaux(['D',4], shape=[1])
                        sage: VM = MinusculeRepresentation(R, CM)
                        sage: CA = crystals.Tableaux(['D',4], shape=[1,1])
                        sage: VA = AdjointRepresentation(R, CA)
                        sage: v = tensor([VM.an_element(), VA.an_element()]); v
                        4*B[[[1]]] # B[[[1], [2]]] + 4*B[[[1]]] # B[[[1], [3]]]
                         + 6*B[[[1]]] # B[[[2], [3]]] + 4*B[[[2]]] # B[[[1], [2]]]
                         + 4*B[[[2]]] # B[[[1], [3]]] + 6*B[[[2]]] # B[[[2], [3]]]
                         + 6*B[[[3]]] # B[[[1], [2]]] + 6*B[[[3]]] # B[[[1], [3]]]
                         + 9*B[[[3]]] # B[[[2], [3]]]
                        sage: v.e(1)  # indirect doctest
                        4*B[[[1]]] # B[[[1], [2]]]
                         + ((4*q+6)/q)*B[[[1]]] # B[[[1], [3]]]
                         + 6*B[[[1]]] # B[[[2], [3]]]
                         + 6*q*B[[[2]]] # B[[[1], [3]]]
                         + 9*B[[[3]]] # B[[[1], [3]]]
                        sage: v.e(2)  # indirect doctest
                        4*B[[[1]]] # B[[[1], [2]]]
                         + ((6*q+4)/q)*B[[[2]]] # B[[[1], [2]]]
                         + 6*B[[[2]]] # B[[[1], [3]]]
                         + 9*B[[[2]]] # B[[[2], [3]]]
                         + 6*q*B[[[3]]] # B[[[1], [2]]]
                        sage: v.e(3)  # indirect doctest
                        0
                        sage: v.e(4)  # indirect doctest
                        0
                    """
                    K_elt = [self._sets[k].K_on_basis(i, elt, -1) for k,elt in enumerate(b)]
                    mon = [self._sets[k].monomial(elt) for k,elt in enumerate(b)]
                    t = self.tensor_constructor(self._sets)
                    ret = self.zero()
                    for k,elt in enumerate(b):
                        ret += t(*(K_elt[:k] + [self._sets[k].e_on_basis(i, elt)] + mon[k+1:]))
                    return ret

                def f_on_basis(self, i, b):
                    r"""
                    Return the action of `f_i` on the basis element
                    indexed by ``b``.

                    INPUT:

                    - ``i`` -- an element of the index set
                    - ``b`` -- an element of basis keys

                    EXAMPLES::

                        sage: from sage.algebras.quantum_groups.representations import \
                        ....:  MinusculeRepresentation, AdjointRepresentation
                        sage: R = ZZ['q'].fraction_field()
                        sage: KM = crystals.KirillovReshetikhin(['B',3,1], 3,1)
                        sage: VM = MinusculeRepresentation(R, KM)
                        sage: KA = crystals.KirillovReshetikhin(['B',3,1], 2,1)
                        sage: VA = AdjointRepresentation(R, KA)
                        sage: v = tensor([VM.an_element(), VA.an_element()]); v
                        4*B[[+++, []]] # B[[]] + 4*B[[+++, []]] # B[[[1], [2]]]
                         + 6*B[[+++, []]] # B[[[1], [3]]] + 4*B[[++-, []]] # B[[]]
                         + 4*B[[++-, []]] # B[[[1], [2]]]
                         + 6*B[[++-, []]] # B[[[1], [3]]] + 6*B[[+-+, []]] # B[[]]
                         + 6*B[[+-+, []]] # B[[[1], [2]]]
                         + 9*B[[+-+, []]] # B[[[1], [3]]]
                        sage: v.f(0)  # indirect doctest
                        ((4*q^4+4)/q^2)*B[[+++, []]] # B[[[1], [2]]]
                         + ((4*q^4+4)/q^2)*B[[++-, []]] # B[[[1], [2]]]
                         + ((6*q^4+6)/q^2)*B[[+-+, []]] # B[[[1], [2]]]
                        sage: v.f(1)  # indirect doctest
                        6*B[[+++, []]] # B[[[2], [3]]]
                         + 6*B[[++-, []]] # B[[[2], [3]]]
                         + 9*B[[+-+, []]] # B[[[2], [3]]]
                         + 6*B[[-++, []]] # B[[]]
                         + 6*B[[-++, []]] # B[[[1], [2]]]
                         + 9*q^2*B[[-++, []]] # B[[[1], [3]]]
                        sage: v.f(2)  # indirect doctest
                        4*B[[+++, []]] # B[[[1], [3]]]
                         + 4*B[[++-, []]] # B[[[1], [3]]]
                         + 4*B[[+-+, []]] # B[[]]
                         + 4*q^2*B[[+-+, []]] # B[[[1], [2]]]
                         + ((6*q^2+6)/q^2)*B[[+-+, []]] # B[[[1], [3]]]
                        sage: v.f(3)  # indirect doctest
                        6*B[[+++, []]] # B[[[1], [0]]]
                         + 4*B[[++-, []]] # B[[]]
                         + 4*B[[++-, []]] # B[[[1], [2]]]
                         + 6*q^2*B[[++-, []]] # B[[[1], [3]]]
                         + 6*B[[++-, []]] # B[[[1], [0]]]
                         + 9*B[[+-+, []]] # B[[[1], [0]]]
                         + 6*B[[+--, []]] # B[[]]
                         + 6*B[[+--, []]] # B[[[1], [2]]]
                         + 9*q^2*B[[+--, []]] # B[[[1], [3]]]
                    """
                    K_elt = [self._sets[k].K_on_basis(i, elt, 1) for k,elt in enumerate(b)]
                    mon = [self._sets[k].monomial(elt) for k,elt in enumerate(b)]
                    t = self.tensor_constructor(self._sets)
                    ret = self.zero()
                    for k,elt in enumerate(b):
                        ret += t(*(mon[:k] + [self._sets[k].f_on_basis(i, elt)] + K_elt[k+1:]))
                    return ret

                def K_on_basis(self, i, b, power=1):
                    r"""
                    Return the action of `K_i` on the basis element indexed by ``b``
                    to the power ``power``.

                    INPUT:

                    - ``i`` -- an element of the index set
                    - ``b`` -- an element of basis keys
                    - ``power`` -- (default: 1) the power of `K_i`

                    EXAMPLES::

                        sage: from sage.algebras.quantum_groups.representations import \
                        ....:  MinusculeRepresentation, AdjointRepresentation
                        sage: R = ZZ['q'].fraction_field()
                        sage: CM = crystals.Tableaux(['A',2], shape=[1])
                        sage: VM = MinusculeRepresentation(R, CM)
                        sage: CA = crystals.Tableaux(['A',2], shape=[2,1])
                        sage: VA = AdjointRepresentation(R, CA)
                        sage: v = tensor([sum(VM.basis()), VA.module_generator()]); v
                        B[[[1]]] # B[[[1, 1], [2]]]
                         + B[[[2]]] # B[[[1, 1], [2]]]
                         + B[[[3]]] # B[[[1, 1], [2]]]
                        sage: v.K(1)  # indirect doctest
                        q^2*B[[[1]]] # B[[[1, 1], [2]]]
                         + B[[[2]]] # B[[[1, 1], [2]]]
                         + q*B[[[3]]] # B[[[1, 1], [2]]]
                        sage: v.K(2, -1)  # indirect doctest
                        1/q*B[[[1]]] # B[[[1, 1], [2]]]
                         + 1/q^2*B[[[2]]] # B[[[1, 1], [2]]]
                         + B[[[3]]] # B[[[1, 1], [2]]]
                    """
                    t = self.tensor_constructor(self._sets)
                    return t(*[self._sets[k].K_on_basis(i, elt, power)
                               for k,elt in enumerate(b)])

        class ParentMethods:
            def tensor(*factors):
                """
                Return the tensor product of ``self`` with the
                representations ``factors``.

                EXAMPLES::

                    sage: from sage.algebras.quantum_groups.representations import \
                    ....:  MinusculeRepresentation, AdjointRepresentation
                    sage: R = ZZ['q'].fraction_field()
                    sage: CM = crystals.Tableaux(['D',4], shape=[1])
                    sage: CA = crystals.Tableaux(['D',4], shape=[1,1])
                    sage: V = MinusculeRepresentation(R, CM)
                    sage: V.tensor(V, V)
                    V((1, 0, 0, 0)) # V((1, 0, 0, 0)) # V((1, 0, 0, 0))
                    sage: A = MinusculeRepresentation(R, CA)
                    sage: V.tensor(A)
                    V((1, 0, 0, 0)) # V((1, 1, 0, 0))
                    sage: B = crystals.Tableaux(['A',2], shape=[1])
                    sage: W = MinusculeRepresentation(R, B)
                    sage: tensor([W,V])
                    Traceback (most recent call last):
                    ...
                    ValueError: all factors must be of the same Cartan type
                    sage: tensor([V,A,W])
                    Traceback (most recent call last):
                    ...
                    ValueError: all factors must be of the same Cartan type
                """
                cartan_type = factors[0].cartan_type()
                if any(V.cartan_type() != cartan_type for V in factors):
                    raise ValueError("all factors must be of the same Cartan type")
                return factors[0].__class__.Tensor(factors, category=tensor.category_from_parents(factors))

        class ElementMethods:
            def e(self, i):
                r"""
                Return the action of `e_i` on ``self``.

                INPUT:

                - ``i`` -- an element of the index set

                EXAMPLES::

                    sage: from sage.algebras.quantum_groups.representations import AdjointRepresentation
                    sage: C = crystals.Tableaux(['G',2], shape=[1,1])
                    sage: R = ZZ['q'].fraction_field()
                    sage: V = AdjointRepresentation(R, C)
                    sage: v = V.an_element(); v
                    2*B[[[1], [2]]] + 2*B[[[1], [3]]] + 3*B[[[2], [3]]]
                    sage: v.e(1)
                    ((3*q^4+3*q^2+3)/q^2)*B[[[1], [3]]]
                    sage: v.e(2)
                    2*B[[[1], [2]]]
                """
                F = self.parent()
                mc = self.monomial_coefficients(copy=False)
                return F.linear_combination((F.e_on_basis(i, m), c)
                                            for m, c in mc.items())

            def f(self, i):
                r"""
                Return the action of `f_i` on ``self``.

                INPUT:

                - ``i`` -- an element of the index set

                EXAMPLES::

                    sage: from sage.algebras.quantum_groups.representations import AdjointRepresentation
                    sage: K = crystals.KirillovReshetikhin(['D',4,1], 2,1)
                    sage: R = ZZ['q'].fraction_field()
                    sage: V = AdjointRepresentation(R, K)
                    sage: v = V.an_element(); v
                    2*B[[]] + 2*B[[[1], [2]]] + 3*B[[[1], [3]]]
                    sage: v.f(0)
                    ((2*q^2+2)/q)*B[[[1], [2]]]
                    sage: v.f(1)
                    3*B[[[2], [3]]]
                    sage: v.f(2)
                    2*B[[[1], [3]]]
                    sage: v.f(3)
                    3*B[[[1], [4]]]
                    sage: v.f(4)
                    3*B[[[1], [-4]]]
                """
                F = self.parent()
                mc = self.monomial_coefficients(copy=False)
                return F.linear_combination((F.f_on_basis(i, m), c)
                                            for m, c in mc.items())

            def K(self, i, power=1):
                r"""
                Return the action of `K_i` on ``self`` to the power ``power``.

                INPUT:

                - ``i`` -- an element of the index set
                - ``power`` -- (default: 1) the power of `K_i`

                EXAMPLES::

                    sage: from sage.algebras.quantum_groups.representations import AdjointRepresentation
                    sage: K = crystals.KirillovReshetikhin(['D',4,2], 1,1)
                    sage: R = ZZ['q'].fraction_field()
                    sage: V = AdjointRepresentation(R, K)
                    sage: v = V.an_element(); v
                    2*B[[]] + 2*B[[[1]]] + 3*B[[[2]]]
                    sage: v.K(0)
                    2*B[[]] + 2/q^2*B[[[1]]] + 3*B[[[2]]]
                    sage: v.K(1)
                    2*B[[]] + 2*q^2*B[[[1]]] + 3/q^2*B[[[2]]]
                    sage: v.K(1, 2)
                    2*B[[]] + 2*q^4*B[[[1]]] + 3/q^4*B[[[2]]]
                    sage: v.K(1, -1)
                    2*B[[]] + 2/q^2*B[[[1]]] + 3*q^2*B[[[2]]]
                """
                F = self.parent()
                mc = self.monomial_coefficients(copy=False)
                return F.linear_combination((F.K_on_basis(i, m, power), c)
                                             for m, c in mc.items())

    class TensorProducts(TensorProductsCategory):
        """
        The category of quantum group representations constructed
        by tensor product of quantum group representations.

        .. WARNING::

            We use the reversed coproduct in order to match the
            tensor product rule on crystals.
        """
        @cached_method
        def extra_super_categories(self):
            """
            EXAMPLES::

                sage: from sage.categories.quantum_group_representations import QuantumGroupRepresentations
                sage: Cat = QuantumGroupRepresentations(ZZ['q'].fraction_field())
                sage: Cat.TensorProducts().extra_super_categories()
                [Category of quantum group representations over
                 Fraction Field of Univariate Polynomial Ring in q over Integer Ring]
            """
            return [self.base_category()]

        class ParentMethods:
            def cartan_type(self):
                """
                Return the Cartan type of ``self``.

                EXAMPLES::

                    sage: from sage.algebras.quantum_groups.representations import MinusculeRepresentation
                    sage: C = crystals.Tableaux(['C',2], shape=[1])
                    sage: R = ZZ['q'].fraction_field()
                    sage: V = MinusculeRepresentation(R, C)
                    sage: T = tensor([V,V])
                    sage: T.cartan_type()
                    ['C', 2]
                """
                return self._sets[0].cartan_type()

    class ParentMethods:
        def _test_representation(self, tester=None, **options):
            """
            Test the quantum group relations on ``self``.

            .. SEEALSO:: :class:`TestSuite`

            EXAMPLES::

                sage: from sage.algebras.quantum_groups.representations import \
                ....:  MinusculeRepresentation, AdjointRepresentation
                sage: C = crystals.Tableaux(['G',2], shape=[1,1])
                sage: R = ZZ['q'].fraction_field()
                sage: V = AdjointRepresentation(R, C)
                sage: V._test_representation()

            We verify that ``C`` does not define a minuscule
            representation::

                sage: M = MinusculeRepresentation(R, C)
                sage: M._test_representation()
                Traceback (most recent call last):
                ...
                AssertionError: [e,f] = (K-K^-1)/(q_i-q_i^-1) -- i: 1 j: 1
            """
            tester = self._tester(**options)
            ct = self.cartan_type()
            d = ct.symmetrizer()
            I = ct.index_set()
            A = ct.cartan_matrix()
            al = ct.root_system().weight_lattice().simple_roots()
            ac = ct.root_system().weight_lattice().simple_coroots()
            q = self.q()
            from sage.algebras.quantum_groups.q_numbers import q_factorial
            def apply_e(d, elt):
                for i in d:
                    elt = elt.e(i)
                return elt

            def apply_f(d, elt):
                for i in d:
                    elt = elt.f(i)
                return elt

            count = 0
            for x in self.basis():
                for i in I:
                    for j in I:
                        tester.assertEqual(x.K(j,-1).f(i).K(j,1),
                                           q**-(al[i].scalar(ac[j]) * d[j]) * x.f(i),
                                           "KfK^-1 -- i: {}, j: {}".format(i,j))
                        tester.assertEqual(x.K(j,-1).e(i).K(j,1),
                                           q**(al[i].scalar(ac[j]) * d[j]) * x.e(i),
                                           "KeK^-1 -- i: {}, j: {}".format(i,j))
                        if i == j:
                            tester.assertEqual(x.f(i).e(i) - x.e(i).f(i),
                                               (x.K(i,1) - x.K(i,-1)) / (q**d[i] - q**(-d[i])),
                                               "[e,f] = (K-K^-1)/(q_i-q_i^-1) -- i: {} j: {}".format(i, j))
                            continue
                        tester.assertEqual(x.f(j).e(i) - x.e(i).f(j), 0,
                                           "[e,f] = 0 -- i: {} j: {}".format(i, j))
                        # Check quantum Serre
                        aij = A[I.index(i),I.index(j)]
                        tester.assertEqual(0,
                                           sum((-1)**n
                                               * q_factorial(1-aij, q**d[i])
                                               / (q_factorial(n, q**d[i])
                                                  * q_factorial(1-aij-n, q**d[i]))
                                               * apply_e([i]*(1-aij-n) + [j] + [i]*n, x)
                                               for n in range(1-aij+1)),
                                           "quantum Serre e -- i: {}, j: {}".format(i,j))
                        tester.assertEqual(0,
                                           sum((-1)**n
                                               * q_factorial(1-aij, q**d[i])
                                               / (q_factorial(n, q**d[i])
                                                  * q_factorial(1-aij-n, q**d[i]))
                                               * apply_f([i]*(1-aij-n) + [j] + [i]*n, x)
                                               for n in range(1-aij+1)),
                                           "quantum Serre f -- i: {}, j: {}".format(i,j))
                count += 1
                if count > tester._max_runs:
                    return

        @abstract_method
        def cartan_type(self):
            """
            Return the Cartan type of ``self``.

            EXAMPLES::

                sage: from sage.algebras.quantum_groups.representations import MinusculeRepresentation
                sage: C = crystals.Tableaux(['C',4], shape=[1])
                sage: R = ZZ['q'].fraction_field()
                sage: V = MinusculeRepresentation(R, C)
                sage: V.cartan_type()
                ['C', 4]
            """

        @cached_method
        def index_set(self):
            """
            Return the index set of ``self``.

            EXAMPLES::

                sage: from sage.algebras.quantum_groups.representations import MinusculeRepresentation
                sage: C = crystals.Tableaux(['C',4], shape=[1])
                sage: R = ZZ['q'].fraction_field()
                sage: V = MinusculeRepresentation(R, C)
                sage: V.index_set()
                (1, 2, 3, 4)
            """
            return self.cartan_type().index_set()

        def q(self):
            r"""
            Return the quantum parameter `q` of ``self``.

            EXAMPLES::

                sage: from sage.algebras.quantum_groups.representations import MinusculeRepresentation
                sage: C = crystals.Tableaux(['C',4], shape=[1])
                sage: R = ZZ['q'].fraction_field()
                sage: V = MinusculeRepresentation(R, C)
                sage: V.q()
                q
            """
            return self._q




"""
Type internals
"""

from cpython.object cimport Py_TYPE, PyTypeObject, Py_TPFLAGS_HEAPTYPE

cdef PyTypeObject* PyInstance_Type = NULL
try:
    from types import InstanceType
    PyInstance_Type = <PyTypeObject*>InstanceType
except ImportError:
    pass


cpdef bint can_assign_class(obj):
    """
    Can we assign ``obj.__class__``?

    Note that Python 3.5 has experimented with allowing assigning
    ``__class__`` in more cases but this was mostly reverted. In this
    function, we apply the Python 2.7 semantics.

    EXAMPLES::

        sage: class X: pass
        sage: from sage.cpython.type import can_assign_class
        sage: can_assign_class(X())
        True
        sage: class Y(int): pass
        sage: from sage.cpython.type import can_assign_class
        sage: can_assign_class(Y())
        True
        sage: can_assign_class(1)
        False
    """
    cdef PyTypeObject* tp = Py_TYPE(obj)
    if tp is PyInstance_Type:
        return True
    return (tp.tp_flags & Py_TPFLAGS_HEAPTYPE) != 0



"IntegerFactorization objects"

from sage.structure.factorization import Factorization
from sage.rings.integer_ring import ZZ


class IntegerFactorization(Factorization):
    """
    A lightweight class for an ``IntegerFactorization`` object,
    inheriting from the more general ``Factorization`` class.

    In the ``Factorization`` class the user has to create a list
    containing the factorization data, which is then passed to the
    actual ``Factorization`` object upon initialization.

    However, for the typical use of integer factorization via
    the ``Integer.factor()`` method in ``sage.rings.integer``
    this is noticeably too much overhead, slowing down the
    factorization of integers of up to about 40 bits by a factor
    of around 10.  Moreover, the initialization done in the
    ``Factorization`` class is typically unnecessary:  the caller
    can guarantee that the list contains pairs of an ``Integer``
    and an ``int``, as well as that the list is sorted.

    AUTHOR:

    - Sebastian Pancratz (2010-01-10)
    """

    def __init__(self, x, unit=None, cr=False, sort=True, simplify=True,
                 unsafe=False):
        """
        Set ``self`` to the factorization object with list ``x``,
        which must be a sorted list of pairs, where each pair contains
        a factor and an exponent.

        If the flag ``unsafe`` is set to ``False`` this method delegates
        the initialization to the parent class, which means that a rather
        lenient and careful way of initialization is chosen.  For example,
        elements are coerced or converted into the right parents, multiple
        occurrences of the same factor are collected (in the commutative
        case), the list is sorted (unless ``sort`` is ``False``) etc.

        However, if the flag is set to ``True``, no error handling is
        carried out.  The list ``x`` is assumed to list of pairs.  The
        type of the factors is assumed to be constant across all factors:
        either ``Integer`` (the generic case) or ``int`` (as supported
        by the flag ``int_`` of the ``factor()`` method).  The type of
        the exponents is assumed to be ``int``.  The list ``x`` itself
        will be referenced in this factorization object and hence the
        caller is responsible for not changing the list after creating
        the factorization.  The unit is assumed to be either ``None`` or
        of type ``Integer``, taking one of the values `+1` or `-1`.

        EXAMPLES::

            sage: factor(15)
            3 * 5

        We check that :trac:`13139` is fixed::

            sage: from sage.structure.factorization_integer import IntegerFactorization
            sage: IntegerFactorization([(3, 1)], unsafe=True)
            3
        """
        if unsafe:
            if unit is None:
                self._Factorization__unit = ZZ._one_element
            else:
                self._Factorization__unit = unit

            self._Factorization__x        = x
            self._Factorization__universe = ZZ
            self._Factorization__cr       = cr

            if sort:
                self.sort()
            if simplify:
                self.simplify()

        else:
            super(IntegerFactorization, self).__init__(x, unit=unit, cr=cr,
                                                       sort=sort,
                                                       simplify=simplify)

    def __sort__(self, key=None):
        """
        Sort the factors in this factorization.

        INPUT:

        - ``key`` -- (default: ``None``) comparison key

        EXAMPLES::

            sage: F = factor(15)
            sage: F.sort(key=lambda x: -x[0])
            sage: F
            5 * 3
        """
        if key is not None:
            self.__x.sort(key=key)
        else:
            self.__x.sort()



r"""
Function fields
"""
#*****************************************************************************
#  Copyright (C) 2005      David Kohel <kohel@maths.usyd.edu>
#                          William Stein <wstein@math.ucsd.edu>
#                2008      Teresa Gomez-Diaz (CNRS) <Teresa.Gomez-Diaz@univ-mlv.fr>
#                2008-2009 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.categories.category import Category
from sage.misc.cachefunc import cached_method
from sage.categories.basic import Fields

class FunctionFields(Category):
    r"""
    The category of function fields.

    EXAMPLES:

    We create the category of function fields::

        sage: C = FunctionFields()
        sage: C
        Category of function fields

    TESTS::

        sage: TestSuite(FunctionFields()).run()
    """
    @cached_method
    def super_categories(self):
        """
        Returns the Category of which this is a direct sub-Category
        For a list off all super categories see all_super_categories

        EXAMPLES::

            sage: FunctionFields().super_categories()
            [Category of fields]
        """
        return[Fields()]

    def _call_(self, x):
        r"""
        Constructs an object in this category from the data in ``x``,
        or throws a TypeError.

        EXAMPLES::

            sage: C = FunctionFields()
            sage: K.<x>=FunctionField(QQ)
            sage: C(K)
            Rational function field in x over Rational Field
            sage: Ky.<y> = K[]
            sage: L = K.extension(y^2-x)
            sage: C(L)
            Function field in y defined by y^2 - x
            sage: C(L.equation_order())
            Function field in y defined by y^2 - x
        """
        try:
            return x.function_field()
        except AttributeError:
            raise  TypeError("unable to canonically associate a function field to %s"%x)

    class ParentMethods:
        pass

    class ElementMethods:
        pass
